from __future__ import annotations
from typing import TYPE_CHECKING, Optional
from .abc import MinecraftPlayer

if TYPE_CHECKING:
    from .hypixel import Hypixel
    from .hypixelfriends import HypixelFriends
    from .guild import Guild
    from .playerstatus import PlayerStatus
    from .player import Player


class HypixelPlayer(MinecraftPlayer):
    """Class for HypixelPlayer-likes to inherit from"""

    _hypy: Hypixel

    async def friends(self) -> HypixelFriends:
        """Gets HypixelFriends of the HypixelPlayer"""
        return await self._hypy.get_friends(str(self.uuid))

    async def status(self) -> PlayerStatus:
        """Gets PlayerStatus of the HypixelPlayer"""
        return await self._hypy.get_player_status(str(self.uuid))

    async def guild(self) -> Optional[Guild]:
        """Gets the Guild of the HypixelPlayer"""
        return await self._hypy.get_guild(playerNameOrUuid=str(self.uuid))

    async def player(self) -> Player:
        """Gets the Player of the HypixelPlayer"""
        return await self._hypy.get_player(nameOrUuid=str(self.uuid))

    async def name(self) -> str:
        """Gets the Name of the HypixelPlayer from the Mojang API"""
        return await self._hypy.mojang.uuid_to_name(str(self.uuid))



class HypyException(Exception):
    """A Hypy Exception"""


class UsernameNotFound(HypyException):
    """Exception to raise when a minecraft user was not found"""

    def __init__(self, username) -> None:
        super().__init__(
            "A minecraft user with the name '{}' could not be found".format(username)
        )
        self.username = username


class UUIDNotFound(HypyException):
    """Exception to raise when a minecraft user was not found"""

    def __init__(self, uuid) -> None:
        super().__init__(
            "A minecraft user with the UUID '{}' could not be found".format(uuid)
        )
        self.uuid = uuid


class InvalidAccountDescriptor(HypyException):
    """Exception to raise when an invalid account descriptor was passed"""

    def __init__(self, account) -> None:
        super().__init__(
            "The passed argument '{}' cannot be an account's name or UUID".format(
                account
            )
        )


class HypixelNoSuccess(HypyException):
    """Exception to raise when Hypixel returns success=false"""

    def __init__(self, cause) -> None:
        super().__init__("Hypixel returned success=false, cause: {}".format(cause))


class NotEnoughArguments(HypyException):
    """Exception to raise when not enough arguments were passed in"""

    def __init__(self, cause) -> None:
        super().__init__(
            "You did not pass enough arguments, at least one of {} is required".format(
                cause
            )
        )


class ApiOffException(HypyException):
    """Exception to raise when a disabled API gets queried"""

    def __init__(self, cause) -> None:
        super().__init__("Api {} is turned off and cant be queried".format(cause))


class MalformedIDException(HypyException):
    """Exception to raise when a malformed Item ID was passed in"""

    def __init__(self, cause) -> None:
        super().__init__("ID {} is malformed".format(cause))


class ContentTypeException(HypyException):
    """Exception to raise when Hypixel API returned the wrong content type"""

    def __init__(self, cause) -> None:
        super().__init__("Hypixel API returned wrong content type: {}".format(cause))


class MissingParamException(HypyException):
    """Exception to raise when parameters are missing"""

    def __init__(self, cause) -> None:
        super().__init__("Missing any parameter from {}".format(cause))


class InvalidSkillException(HypyException):
    """Exception to raise when an invalid skill was passed in"""

    def __init__(self, cause) -> None:
        super().__init__("The skill {} is invalid".format(cause))


class InvalidProfileException(HypyException):
    """Exception to raise when a profile was not found"""

    def __init__(self, cause) -> None:
        super().__init__("The profile {} is invalid".format(cause))


class InvalidProfileNameException(HypyException):
    """Exception to raise when an invalid profile name was passed in"""

    def __init__(self, cause) -> None:
        super().__init__("The profile name {} is invalid".format(cause))


class ExceededMaxRetries(HypyException):
    """Exception to raise when max retries were exceeded"""

    def __init__(self, cause) -> None:
        super().__init__("Exceeded max retries: {}".format(str(cause)))


class InvalidQueryTypeException(HypyException):
    """Exception to raise when an invalid query type was passed in"""

    def __init__(self, cause) -> None:
        super().__init__("Invalid query type: {}".format(cause))


class NoProfilesFoundException(HypyException):
    """Exception to raise when no profiles were found"""

    def __init__(self, cause) -> None:
        super().__init__("Couldnt find any profiles for uuid: {}".format(cause))


class NoProfileInfoAvailableException(HypyException):
    """Exception to raise when no profiles provide any skill information"""

    def __init__(self) -> None:
        super().__init__(
            "None of the available profiles had any of the information needed to find profile"
        )


class NotInGuildException(HypyException):
    """Exception to raise a person is not in guild but guild is queried for person"""  # this docstring sucks?

    def __init__(self, cause) -> None:
        super().__init__("{} is not in the guild".format(cause))


class MaroNoSuccess(HypyException):
    """Exception to raise when the Maro API returns success=false"""

    def __init__(self, status, cause) -> None:
        super().__init__("Maro returned status={}, cause: {}".format(status, cause))



from typing import Union, List
from enum import Enum, auto
from .auction import Auction
from .uuid import UUID
from .hypyobject import HypyObject


class FilterType(Enum):
    """Types to filter Auctions by"""

    NAME = auto()
    AUCTIONEER = auto()
    LORE = auto()


class SkyblockAuctions(HypyObject):
    """SkyBlock Auction data"""

    __slots__ = ("_raw", "_auctions", "_hypy", "_num")

    def __init__(self, data, auc, hypy) -> None:
        self._raw = data
        self._auctions = auc
        self._num = len(data)
        self._hypy = hypy

    def __len__(self) -> int:
        return self._num

    def __getitem__(self, key) -> Auction:
        return self._auctions[key]

    async def find_auctions(
        self,
        by: Union[str, FilterType] = "name",
        query: str = "",
        case_sensitive: bool = True,
    ) -> List[Auction]:
        """Find Auctions

        :param by: What to filter by
        :param query: The query
        :param case_sensitive: Whether searching should be case sensitive
        """
        if not case_sensitive:
            query = query.lower()
        if isinstance(by, str):
            by = FilterType[str(by).upper()]
        if by == FilterType.NAME:
            return [
                auction
                for auction in self._auctions
                if query
                in (auction.item_name if case_sensitive else auction.item_name.lower())
            ]
        if by == FilterType.AUCTIONEER:
            auc_uuid = UUID(await self._hypy.mojang.name_to_uuid(query)).no_dashes
            return [
                auction
                for auction in self._auctions
                if auc_uuid == auction.auctioneer_uuid.no_dashes
            ]
        if by == FilterType.LORE:
            return [
                auction
                for auction in self._auctions
                if query
                in (
                    auction.safe_item_lore
                    if case_sensitive
                    else auction.safe_item_lore.lower()
                )
            ]
        else:
            return []



from orjson import orjson
import aiofiles


class HypyObject:
    """A base object for API Data"""

    _raw: dict

    async def save_json(self, filename) -> None:
        """Save raw API data to disk

        :param filename: The name of the file to save to
        """
        async with aiofiles.open(filename, mode="w+") as f:
            await f.write(orjson.dumps(self._raw, indent=4))



import re
from typing import Union, Tuple
from io import BytesIO
from base64 import b64decode
from datetime import datetime
from nbt.nbt import TAG_Compound, TAG_List
from nbt import nbt
from .vars import (
    xp_cumulative,
    cata_xp_cumulative,
    cosmetic_xp_cumulative,
    item_id_table,
)
from .hypixelresources import SkyBlockResources
from .exceptions import MalformedIDException

uuid_regex = re.compile(
    r"[0-9A-Za-z]{8}-?[0-9A-Za-z]{4}-?4[0-9A-Za-z]{3}-?[89ABab][0-9A-Za-z]{3}-?[0-9A-Za-z]{12}"
)

DATETIME_FORMAT = r"%m/%d/%y %I:%M %p"
AUC_DATETIME_FORMAT = r"%d/%m/%y %H:%M"

LORE_REGEX = re.compile(r"§.")


def get_auction_lores(auctions):
    """return lore for each auction, use asyncio.to_thread for this"""
    return [get_safe_content(i["item_lore"]) for i in auctions]


def get_safe_content(content) -> str:
    """get safe content"""
    return LORE_REGEX.sub("", content)


def is_uuid(uuid: str) -> bool:
    """Check if parameter is a UUID"""
    # maybe a better function name? could be misleading
    return bool(re.fullmatch(uuid_regex, uuid))


username_regex = r"\w{3,16}"


def is_username(username) -> bool:
    """Check if parameter is a username"""
    # maybe a better function name? could be misleading
    return bool(re.fullmatch(username_regex, username))


def lvlCalc(xp: Union[int, float], skilltype: str) -> float:
    """calculate level for any given skill"""
    xp = int(xp)
    min_lvl = -1
    rel_xp_cumulative = xp_cumulative
    if skilltype in ["catacombs"]:
        rel_xp_cumulative = cata_xp_cumulative
    if skilltype in ["social", "runecrafting"]:
        rel_xp_cumulative = cosmetic_xp_cumulative
    for i in range(62):
        if i not in rel_xp_cumulative:
            rel_xp_cumulative[i] = 11167242500000000000000000000  # ridic high number
    if xp is None:
        return 0.0
    for _, xp_req in rel_xp_cumulative.items():
        if xp_req < xp:
            min_lvl += 1
    if min_lvl < 0:
        return 0.0
    xp_diff = rel_xp_cumulative[min_lvl + 1] - rel_xp_cumulative[min_lvl]
    xp_to_next = rel_xp_cumulative[min_lvl + 1] - xp
    min_lvl_fraction = 1 - (xp_to_next / xp_diff)
    real_level = min_lvl + min_lvl_fraction
    # If skill is level 50, note that and continue (yes im very lazy)
    if real_level > 50:
        if skilltype not in ["enchanting", "farming", "mining"]:
            return 50.0
    if real_level > 60:
        return 60.0
    return float(real_level)


def item_id_to_name(item_id) -> Tuple[str, str]:
    """Return name for given item id"""
    default = "stone"
    default_display = "Error parsing lol"
    i_pair = str(item_id).split(":")
    if len(i_pair) == 1:
        item_ids = [x for x in item_id_table if x["type"] == int(i_pair[0])]
        if len(item_ids) == 0:
            return default, default_display
        return str(item_ids[0]["text_type"]), str(item_ids[0]["name"])
    elif len(i_pair) == 2:
        i_list = [x for x in item_id_table if x["type"] == int(i_pair[0])]
        if len(i_list) == 0:
            return default, default_display
        i_listm = [x for x in i_list if x["meta"] == int(i_pair[1])]
        if len(i_listm) == 0:
            return default, default_display
        return str(i_listm[0]["text_type"]), str(i_listm[0]["name"])
    else:
        raise MalformedIDException(item_id)


def decode_nbt(raw_data):
    """Decode NBT data"""
    nbtdata = nbt.NBTFile(fileobj=BytesIO(b64decode(raw_data)))
    return nbtdata


# https://github.com/twoolie/NBT/blob/master/examples/utilities.py
def unpack_nbt(tag):
    """
    Unpack an NBT tag into a native Python data structure.
    """

    if isinstance(tag, TAG_List):
        return [unpack_nbt(i) for i in tag.tags]
    elif isinstance(tag, TAG_Compound):
        return dict((i.name, unpack_nbt(i)) for i in tag.tags)
    else:
        return tag.value


class Utils:
    """Some misc functions used in hypy"""

    def __init__(self, resources: SkyBlockResources):
        self.resources = resources

    def lvlCalc(self, xp, skilltype) -> float:
        """calculate level for any given skill using cumulative xp from the api"""
        xp = int(xp)
        min_lvl = -1
        if xp is None:
            return 0
        rel_xp_cumulative_data = self.resources.cumulative_xp[skilltype]
        rel_xp_cumulative = rel_xp_cumulative_data["levels"]
        for _, xp_req in rel_xp_cumulative.items():
            if xp_req < xp:
                min_lvl += 1
        if min_lvl < 0:
            return 0
        if min_lvl >= rel_xp_cumulative_data["details"]["max_level"]:
            return rel_xp_cumulative_data["details"]["max_level"]
        xp_diff = rel_xp_cumulative[min_lvl + 1] - rel_xp_cumulative[min_lvl]
        xp_to_next = rel_xp_cumulative[min_lvl + 1] - xp
        min_lvl_fraction = 1 - (xp_to_next / xp_diff)
        real_level = min_lvl + min_lvl_fraction
        # If skill is level 50, note that and continue (yes im very lazy)
        if real_level > rel_xp_cumulative_data["details"]["max_level"]:
            return rel_xp_cumulative_data["details"]["max_level"]
        return float(real_level)


def parse_timestamp(to_parse: str) -> datetime:
    """Return datetime for given timestamp"""
    if any([x in to_parse for x in ["AM", "PM"]]):
        dt = datetime.strptime(to_parse, DATETIME_FORMAT)
    else:
        dt = datetime.strptime(to_parse, AUC_DATETIME_FORMAT)
    return dt
    # dt = dt.replace(tzinfo=pytz.timezone('America/New_York'))
    # print(dt)
    ##print(datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, tz=pytz.timezone('America/New_York')).strftime(datetime_format))
    # exit(0)
    # return dt.astimezone(tz=None)



# these are all only used by the static version of lvlCalc (which you shouldnt use!)

cata_xp_cumulative = {
    0: 0,
    1: 50,
    2: 125,
    3: 235,
    4: 395,
    5: 625,
    6: 955,
    7: 1425,
    8: 2095,
    9: 3045,
    10: 4385,
    11: 6275,
    12: 8940,
    13: 12700,
    14: 17960,
    15: 25340,
    16: 35640,
    17: 50040,
    18: 70040,
    19: 97640,
    20: 135640,
    21: 188140,
    22: 259640,
    23: 356640,
    24: 488640,
    25: 668640,
    26: 911640,
    27: 1239640,
    28: 1684640,
    29: 2284640,
    30: 3084640,
    31: 4149640,
    32: 5559640,
    33: 7459640,
    34: 9959640,
    35: 13259640,
    36: 17559640,
    37: 23159640,
    38: 30359640,
    39: 39559640,
    40: 51559640,
    41: 66559640,
    42: 85559640,
    43: 109559640,
    44: 139559640,
    45: 177559640,
    46: 225559640,
    47: 285559640,
    48: 360559640,
    49: 453559640,
    50: 569809640,
}

xp_cumulative = {
    0: 0,
    1: 50,
    2: 175,
    3: 375,
    4: 675,
    5: 1175,
    6: 1925,
    7: 2925,
    8: 4425,
    9: 6425,
    10: 9925,
    11: 14925,
    12: 22425,
    13: 32425,
    14: 47425,
    15: 67425,
    16: 97425,
    17: 147425,
    18: 222425,
    19: 322425,
    20: 522425,
    21: 822425,
    22: 1222425,
    23: 1722425,
    24: 2322425,
    25: 3022425,
    26: 3822425,
    27: 4722425,
    28: 5722425,
    29: 6822425,
    30: 8022425,
    31: 9322425,
    32: 10722425,
    33: 12222425,
    34: 13822425,
    35: 15522425,
    36: 17322425,
    37: 19222425,
    38: 21222425,
    39: 23322425,
    40: 25522425,
    41: 27822425,
    42: 30222425,
    43: 32722425,
    44: 35322425,
    45: 38072425,
    46: 40972425,
    47: 44072425,
    48: 47472425,
    49: 51172425,
    50: 55172425,
    51: 59472425,
    52: 64072425,
    53: 68972425,
    54: 74172425,
    55: 79672425,
    56: 85472425,
    57: 91572425,
    58: 97972425,
    59: 104672425,
    60: 111672425,
}

cosmetic_xp_cumulative = {
    0: 0,
    1: 50,
    2: 150,
    3: 275,
    4: 435,
    5: 635,
    6: 885,
    7: 1200,
    8: 1600,
    9: 2100,
    10: 2725,
    11: 3510,
    12: 4510,
    13: 5760,
    14: 7325,
    15: 9325,
    16: 11825,
    17: 14950,
    18: 18950,
    19: 23950,
    20: 30200,
    21: 38050,
    22: 47850,
    23: 60100,
    24: 75400,
    25: 94450,
}

types = {
    "fishing": "angler",
    "alchemy": "concoctor",
    "enchanting": "augmentation",
    "combat": "combat",
    "taming": "domesticator",
    "mining": "excavator",
    "foraging": "gatherer",
    "farming": "harvester",
    "catacombs": "dungeoneer",
}

skills = [
    "fishing",
    "alchemy",
    "taming",
    "enchanting",
    "combat",
    "mining",
    "farming",
    "foraging",
    "catacombs",
]
havg_skills = [
    "fishing",
    "alchemy",
    "enchanting",
    "combat",
    "mining",
    "farming",
    "foraging",
    "catacombs",
]

cute_names = [
    "Apple",
    "Banana",
    "Blueberry",
    "Coconut",
    "Cucumber",
    "Grapes",
    "Kiwi",
    "Lemon",
    "Lime",
    "Mango",
    "Orange",
    "Papaya",
    "Peach",
    "Pear",
    "Pineapple",
    "Pomegranate",
    "Raspberry",
    "Strawberry",
    "Tomato",
    "Watermelon",
    "Zucchini",
]

slayer = ["wolf", "spider", "zombie", "enderman"]

misc_skills = ["carpentry", "runecrafting"]

classes = ["healer", "berserk", "archer", "mage", "tank"]

item_id_table = [
    {"type": 0, "meta": 0, "name": "Air", "text_type": "air"},
    {"type": 1, "meta": 0, "name": "Stone", "text_type": "stone"},
    {"type": 1, "meta": 1, "name": "Granite", "text_type": "stone"},
    {"type": 1, "meta": 2, "name": "Polished Granite", "text_type": "stone"},
    {"type": 1, "meta": 3, "name": "Diorite", "text_type": "stone"},
    {"type": 1, "meta": 4, "name": "Polished Diorite", "text_type": "stone"},
    {"type": 1, "meta": 5, "name": "Andesite", "text_type": "stone"},
    {"type": 1, "meta": 6, "name": "Polished Andesite", "text_type": "stone"},
    {"type": 2, "meta": 0, "name": "Grass", "text_type": "grass"},
    {"type": 3, "meta": 0, "name": "Dirt", "text_type": "dirt"},
    {"type": 3, "meta": 1, "name": "Coarse Dirt", "text_type": "dirt"},
    {"type": 3, "meta": 2, "name": "Podzol", "text_type": "dirt"},
    {"type": 4, "meta": 0, "name": "Cobblestone", "text_type": "cobblestone"},
    {"type": 5, "meta": 0, "name": "Oak Wood Plank", "text_type": "planks"},
    {"type": 5, "meta": 1, "name": "Spruce Wood Plank", "text_type": "planks"},
    {"type": 5, "meta": 2, "name": "Birch Wood Plank", "text_type": "planks"},
    {"type": 5, "meta": 3, "name": "Jungle Wood Plank", "text_type": "planks"},
    {"type": 5, "meta": 4, "name": "Acacia Wood Plank", "text_type": "planks"},
    {"type": 5, "meta": 5, "name": "Dark Oak Wood Plank", "text_type": "planks"},
    {"type": 6, "meta": 0, "name": "Oak Sapling", "text_type": "sapling"},
    {"type": 6, "meta": 1, "name": "Spruce Sapling", "text_type": "sapling"},
    {"type": 6, "meta": 2, "name": "Birch Sapling", "text_type": "sapling"},
    {"type": 6, "meta": 3, "name": "Jungle Sapling", "text_type": "sapling"},
    {"type": 6, "meta": 4, "name": "Acacia Sapling", "text_type": "sapling"},
    {"type": 6, "meta": 5, "name": "Dark Oak Sapling", "text_type": "sapling"},
    {"type": 7, "meta": 0, "name": "Bedrock", "text_type": "bedrock"},
    {"type": 8, "meta": 0, "name": "Flowing Water", "text_type": "flowing_water"},
    {"type": 9, "meta": 0, "name": "Still Water", "text_type": "water"},
    {"type": 10, "meta": 0, "name": "Flowing Lava", "text_type": "flowing_lava"},
    {"type": 11, "meta": 0, "name": "Still Lava", "text_type": "lava"},
    {"type": 12, "meta": 0, "name": "Sand", "text_type": "sand"},
    {"type": 12, "meta": 1, "name": "Red Sand", "text_type": "sand"},
    {"type": 13, "meta": 0, "name": "Gravel", "text_type": "gravel"},
    {"type": 14, "meta": 0, "name": "Gold Ore", "text_type": "gold_ore"},
    {"type": 15, "meta": 0, "name": "Iron Ore", "text_type": "iron_ore"},
    {"type": 16, "meta": 0, "name": "Coal Ore", "text_type": "coal_ore"},
    {"type": 17, "meta": 0, "name": "Oak Wood", "text_type": "log"},
    {"type": 17, "meta": 1, "name": "Spruce Wood", "text_type": "log"},
    {"type": 17, "meta": 2, "name": "Birch Wood", "text_type": "log"},
    {"type": 17, "meta": 3, "name": "Jungle Wood", "text_type": "log"},
    {"type": 18, "meta": 0, "name": "Oak Leaves", "text_type": "leaves"},
    {"type": 18, "meta": 1, "name": "Spruce Leaves", "text_type": "leaves"},
    {"type": 18, "meta": 2, "name": "Birch Leaves", "text_type": "leaves"},
    {"type": 18, "meta": 3, "name": "Jungle Leaves", "text_type": "leaves"},
    {"type": 19, "meta": 0, "name": "Sponge", "text_type": "sponge"},
    {"type": 19, "meta": 1, "name": "Wet Sponge", "text_type": "sponge"},
    {"type": 20, "meta": 0, "name": "Glass", "text_type": "glass"},
    {"type": 21, "meta": 0, "name": "Lapis Lazuli Ore", "text_type": "lapis_ore"},
    {"type": 22, "meta": 0, "name": "Lapis Lazuli Block", "text_type": "lapis_block"},
    {"type": 23, "meta": 0, "name": "Dispenser", "text_type": "dispenser"},
    {"type": 24, "meta": 0, "name": "Sandstone", "text_type": "sandstone"},
    {"type": 24, "meta": 1, "name": "Chiseled Sandstone", "text_type": "sandstone"},
    {"type": 24, "meta": 2, "name": "Smooth Sandstone", "text_type": "sandstone"},
    {"type": 25, "meta": 0, "name": "Note Block", "text_type": "noteblock"},
    {"type": 26, "meta": 0, "name": "Bed", "text_type": "bed"},
    {"type": 27, "meta": 0, "name": "Powered Rail", "text_type": "golden_rail"},
    {"type": 28, "meta": 0, "name": "Detector Rail", "text_type": "detector_rail"},
    {"type": 29, "meta": 0, "name": "Sticky Piston", "text_type": "sticky_piston"},
    {"type": 30, "meta": 0, "name": "Cobweb", "text_type": "web"},
    {"type": 31, "meta": 0, "name": "Dead Shrub", "text_type": "tallgrass"},
    {"type": 31, "meta": 1, "name": "Grass", "text_type": "tallgrass"},
    {"type": 31, "meta": 2, "name": "Fern", "text_type": "tallgrass"},
    {"type": 32, "meta": 0, "name": "Dead Bush", "text_type": "deadbush"},
    {"type": 33, "meta": 0, "name": "Piston", "text_type": "piston"},
    {"type": 34, "meta": 0, "name": "Piston Head", "text_type": "piston_head"},
    {"type": 35, "meta": 0, "name": "White Wool", "text_type": "wool"},
    {"type": 35, "meta": 1, "name": "Orange Wool", "text_type": "wool"},
    {"type": 35, "meta": 2, "name": "Magenta Wool", "text_type": "wool"},
    {"type": 35, "meta": 3, "name": "Light Blue Wool", "text_type": "wool"},
    {"type": 35, "meta": 4, "name": "Yellow Wool", "text_type": "wool"},
    {"type": 35, "meta": 5, "name": "Lime Wool", "text_type": "wool"},
    {"type": 35, "meta": 6, "name": "Pink Wool", "text_type": "wool"},
    {"type": 35, "meta": 7, "name": "Gray Wool", "text_type": "wool"},
    {"type": 35, "meta": 8, "name": "Light Gray Wool", "text_type": "wool"},
    {"type": 35, "meta": 9, "name": "Cyan Wool", "text_type": "wool"},
    {"type": 35, "meta": 10, "name": "Purple Wool", "text_type": "wool"},
    {"type": 35, "meta": 11, "name": "Blue Wool", "text_type": "wool"},
    {"type": 35, "meta": 12, "name": "Brown Wool", "text_type": "wool"},
    {"type": 35, "meta": 13, "name": "Green Wool", "text_type": "wool"},
    {"type": 35, "meta": 14, "name": "Red Wool", "text_type": "wool"},
    {"type": 35, "meta": 15, "name": "Black Wool", "text_type": "wool"},
    {"type": 37, "meta": 0, "name": "Dandelion", "text_type": "yellow_flower"},
    {"type": 38, "meta": 0, "name": "Poppy", "text_type": "red_flower"},
    {"type": 38, "meta": 1, "name": "Blue Orchid", "text_type": "red_flower"},
    {"type": 38, "meta": 2, "name": "Allium", "text_type": "red_flower"},
    {"type": 38, "meta": 3, "name": "Azure Bluet", "text_type": "red_flower"},
    {"type": 38, "meta": 4, "name": "Red Tulip", "text_type": "red_flower"},
    {"type": 38, "meta": 5, "name": "Orange Tulip", "text_type": "red_flower"},
    {"type": 38, "meta": 6, "name": "White Tulip", "text_type": "red_flower"},
    {"type": 38, "meta": 7, "name": "Pink Tulip", "text_type": "red_flower"},
    {"type": 38, "meta": 8, "name": "Oxeye Daisy", "text_type": "red_flower"},
    {"type": 39, "meta": 0, "name": "Brown Mushroom", "text_type": "brown_mushroom"},
    {"type": 40, "meta": 0, "name": "Red Mushroom", "text_type": "red_mushroom"},
    {"type": 41, "meta": 0, "name": "Gold Block", "text_type": "gold_block"},
    {"type": 42, "meta": 0, "name": "Iron Block", "text_type": "iron_block"},
    {
        "type": 43,
        "meta": 0,
        "name": "Double Stone Slab",
        "text_type": "double_stone_slab",
    },
    {
        "type": 43,
        "meta": 1,
        "name": "Double Sandstone Slab",
        "text_type": "double_stone_slab",
    },
    {
        "type": 43,
        "meta": 2,
        "name": "Double Wooden Slab",
        "text_type": "double_stone_slab",
    },
    {
        "type": 43,
        "meta": 3,
        "name": "Double Cobblestone Slab",
        "text_type": "double_stone_slab",
    },
    {
        "type": 43,
        "meta": 4,
        "name": "Double Brick Slab",
        "text_type": "double_stone_slab",
    },
    {
        "type": 43,
        "meta": 5,
        "name": "Double Stone Brick Slab",
        "text_type": "double_stone_slab",
    },
    {
        "type": 43,
        "meta": 6,
        "name": "Double Nether Brick Slab",
        "text_type": "double_stone_slab",
    },
    {
        "type": 43,
        "meta": 7,
        "name": "Double Quartz Slab",
        "text_type": "double_stone_slab",
    },
    {"type": 44, "meta": 0, "name": "Stone Slab", "text_type": "stone_slab"},
    {"type": 44, "meta": 1, "name": "Sandstone Slab", "text_type": "stone_slab"},
    {"type": 44, "meta": 2, "name": "Wooden Slab", "text_type": "stone_slab"},
    {"type": 44, "meta": 3, "name": "Cobblestone Slab", "text_type": "stone_slab"},
    {"type": 44, "meta": 4, "name": "Brick Slab", "text_type": "stone_slab"},
    {"type": 44, "meta": 5, "name": "Stone Brick Slab", "text_type": "stone_slab"},
    {"type": 44, "meta": 6, "name": "Nether Brick Slab", "text_type": "stone_slab"},
    {"type": 44, "meta": 7, "name": "Quartz Slab", "text_type": "stone_slab"},
    {"type": 45, "meta": 0, "name": "Bricks", "text_type": "brick_block"},
    {"type": 46, "meta": 0, "name": "TNT", "text_type": "tnt"},
    {"type": 47, "meta": 0, "name": "Bookshelf", "text_type": "bookshelf"},
    {"type": 48, "meta": 0, "name": "Moss Stone", "text_type": "mossy_cobblestone"},
    {"type": 49, "meta": 0, "name": "Obsidian", "text_type": "obsidian"},
    {"type": 50, "meta": 0, "name": "Torch", "text_type": "torch"},
    {"type": 51, "meta": 0, "name": "Fire", "text_type": "fire"},
    {"type": 52, "meta": 0, "name": "Monster Spawner", "text_type": "mob_spawner"},
    {"type": 53, "meta": 0, "name": "Oak Wood Stairs", "text_type": "oak_stairs"},
    {"type": 54, "meta": 0, "name": "Chest", "text_type": "chest"},
    {"type": 55, "meta": 0, "name": "Redstone Wire", "text_type": "redstone_wire"},
    {"type": 56, "meta": 0, "name": "Diamond Ore", "text_type": "diamond_ore"},
    {"type": 57, "meta": 0, "name": "Diamond Block", "text_type": "diamond_block"},
    {"type": 58, "meta": 0, "name": "Crafting Table", "text_type": "crafting_table"},
    {"type": 59, "meta": 0, "name": "Wheat Crops", "text_type": "wheat"},
    {"type": 60, "meta": 0, "name": "Farmland", "text_type": "farmland"},
    {"type": 61, "meta": 0, "name": "Furnace", "text_type": "furnace"},
    {"type": 62, "meta": 0, "name": "Burning Furnace", "text_type": "lit_furnace"},
    {
        "type": 63,
        "meta": 0,
        "name": "Standing Sign Block",
        "text_type": "standing_sign",
    },
    {"type": 64, "meta": 0, "name": "Oak Door Block", "text_type": "wooden_door"},
    {"type": 65, "meta": 0, "name": "Ladder", "text_type": "ladder"},
    {"type": 66, "meta": 0, "name": "Rail", "text_type": "rail"},
    {"type": 67, "meta": 0, "name": "Cobblestone Stairs", "text_type": "stone_stairs"},
    {
        "type": 68,
        "meta": 0,
        "name": "Wall-mounted Sign Block",
        "text_type": "wall_sign",
    },
    {"type": 69, "meta": 0, "name": "Lever", "text_type": "lever"},
    {
        "type": 70,
        "meta": 0,
        "name": "Stone Pressure Plate",
        "text_type": "stone_pressure_plate",
    },
    {"type": 71, "meta": 0, "name": "Iron Door Block", "text_type": "iron_door"},
    {
        "type": 72,
        "meta": 0,
        "name": "Wooden Pressure Plate",
        "text_type": "wooden_pressure_plate",
    },
    {"type": 73, "meta": 0, "name": "Redstone Ore", "text_type": "redstone_ore"},
    {
        "type": 74,
        "meta": 0,
        "name": "Glowing Redstone Ore",
        "text_type": "lit_redstone_ore",
    },
    {
        "type": 75,
        "meta": 0,
        "name": "Redstone Torch (off)",
        "text_type": "unlit_redstone_torch",
    },
    {
        "type": 76,
        "meta": 0,
        "name": "Redstone Torch (on)",
        "text_type": "redstone_torch",
    },
    {"type": 77, "meta": 0, "name": "Stone Button", "text_type": "stone_button"},
    {"type": 78, "meta": 0, "name": "Snow", "text_type": "snow_layer"},
    {"type": 79, "meta": 0, "name": "Ice", "text_type": "ice"},
    {"type": 80, "meta": 0, "name": "Snow Block", "text_type": "snow"},
    {"type": 81, "meta": 0, "name": "Cactus", "text_type": "cactus"},
    {"type": 82, "meta": 0, "name": "Clay", "text_type": "clay"},
    {"type": 83, "meta": 0, "name": "Sugar Canes", "text_type": "reeds"},
    {"type": 84, "meta": 0, "name": "Jukebox", "text_type": "jukebox"},
    {"type": 85, "meta": 0, "name": "Oak Fence", "text_type": "fence"},
    {"type": 86, "meta": 0, "name": "Pumpkin", "text_type": "pumpkin"},
    {"type": 87, "meta": 0, "name": "Netherrack", "text_type": "netherrack"},
    {"type": 88, "meta": 0, "name": "Soul Sand", "text_type": "soul_sand"},
    {"type": 89, "meta": 0, "name": "Glowstone", "text_type": "glowstone"},
    {"type": 90, "meta": 0, "name": "Nether Portal", "text_type": "portal"},
    {"type": 91, "meta": 0, "name": "Jack o'Lantern", "text_type": "lit_pumpkin"},
    {"type": 92, "meta": 0, "name": "Cake Block", "text_type": "cake"},
    {
        "type": 93,
        "meta": 0,
        "name": "Redstone Repeater Block (off)",
        "text_type": "unpowered_repeater",
    },
    {
        "type": 94,
        "meta": 0,
        "name": "Redstone Repeater Block (on)",
        "text_type": "powered_repeater",
    },
    {
        "type": 95,
        "meta": 0,
        "name": "White Stained Glass",
        "text_type": "stained_glass",
    },
    {
        "type": 95,
        "meta": 1,
        "name": "Orange Stained Glass",
        "text_type": "stained_glass",
    },
    {
        "type": 95,
        "meta": 2,
        "name": "Magenta Stained Glass",
        "text_type": "stained_glass",
    },
    {
        "type": 95,
        "meta": 3,
        "name": "Light Blue Stained Glass",
        "text_type": "stained_glass",
    },
    {
        "type": 95,
        "meta": 4,
        "name": "Yellow Stained Glass",
        "text_type": "stained_glass",
    },
    {"type": 95, "meta": 5, "name": "Lime Stained Glass", "text_type": "stained_glass"},
    {"type": 95, "meta": 6, "name": "Pink Stained Glass", "text_type": "stained_glass"},
    {"type": 95, "meta": 7, "name": "Gray Stained Glass", "text_type": "stained_glass"},
    {
        "type": 95,
        "meta": 8,
        "name": "Light Gray Stained Glass",
        "text_type": "stained_glass",
    },
    {"type": 95, "meta": 9, "name": "Cyan Stained Glass", "text_type": "stained_glass"},
    {
        "type": 95,
        "meta": 10,
        "name": "Purple Stained Glass",
        "text_type": "stained_glass",
    },
    {
        "type": 95,
        "meta": 11,
        "name": "Blue Stained Glass",
        "text_type": "stained_glass",
    },
    {
        "type": 95,
        "meta": 12,
        "name": "Brown Stained Glass",
        "text_type": "stained_glass",
    },
    {
        "type": 95,
        "meta": 13,
        "name": "Green Stained Glass",
        "text_type": "stained_glass",
    },
    {"type": 95, "meta": 14, "name": "Red Stained Glass", "text_type": "stained_glass"},
    {
        "type": 95,
        "meta": 15,
        "name": "Black Stained Glass",
        "text_type": "stained_glass",
    },
    {"type": 96, "meta": 0, "name": "Wooden Trapdoor", "text_type": "trapdoor"},
    {"type": 97, "meta": 0, "name": "Stone Monster Egg", "text_type": "monster_egg"},
    {
        "type": 97,
        "meta": 1,
        "name": "Cobblestone Monster Egg",
        "text_type": "monster_egg",
    },
    {
        "type": 97,
        "meta": 2,
        "name": "Stone Brick Monster Egg",
        "text_type": "monster_egg",
    },
    {
        "type": 97,
        "meta": 3,
        "name": "Mossy Stone Brick Monster Egg",
        "text_type": "monster_egg",
    },
    {
        "type": 97,
        "meta": 4,
        "name": "Cracked Stone Brick Monster Egg",
        "text_type": "monster_egg",
    },
    {
        "type": 97,
        "meta": 5,
        "name": "Chiseled Stone Brick Monster Egg",
        "text_type": "monster_egg",
    },
    {"type": 98, "meta": 0, "name": "Stone Bricks", "text_type": "stonebrick"},
    {"type": 98, "meta": 1, "name": "Mossy Stone Bricks", "text_type": "stonebrick"},
    {"type": 98, "meta": 2, "name": "Cracked Stone Bricks", "text_type": "stonebrick"},
    {"type": 98, "meta": 3, "name": "Chiseled Stone Bricks", "text_type": "stonebrick"},
    {
        "type": 99,
        "meta": 0,
        "name": "Brown Mushroom Block",
        "text_type": "brown_mushroom_block",
    },
    {
        "type": 100,
        "meta": 0,
        "name": "Red Mushroom Block",
        "text_type": "red_mushroom_block",
    },
    {"type": 101, "meta": 0, "name": "Iron Bars", "text_type": "iron_bars"},
    {"type": 102, "meta": 0, "name": "Glass Pane", "text_type": "glass_pane"},
    {"type": 103, "meta": 0, "name": "Melon Block", "text_type": "melon_block"},
    {"type": 104, "meta": 0, "name": "Pumpkin Stem", "text_type": "pumpkin_stem"},
    {"type": 105, "meta": 0, "name": "Melon Stem", "text_type": "melon_stem"},
    {"type": 106, "meta": 0, "name": "Vines", "text_type": "vine"},
    {"type": 107, "meta": 0, "name": "Oak Fence Gate", "text_type": "fence_gate"},
    {"type": 108, "meta": 0, "name": "Brick Stairs", "text_type": "brick_stairs"},
    {
        "type": 109,
        "meta": 0,
        "name": "Stone Brick Stairs",
        "text_type": "stone_brick_stairs",
    },
    {"type": 110, "meta": 0, "name": "Mycelium", "text_type": "mycelium"},
    {"type": 111, "meta": 0, "name": "Lily Pad", "text_type": "waterlily"},
    {"type": 112, "meta": 0, "name": "Nether Brick", "text_type": "nether_brick"},
    {
        "type": 113,
        "meta": 0,
        "name": "Nether Brick Fence",
        "text_type": "nether_brick_fence",
    },
    {
        "type": 114,
        "meta": 0,
        "name": "Nether Brick Stairs",
        "text_type": "nether_brick_stairs",
    },
    {"type": 115, "meta": 0, "name": "Nether Wart", "text_type": "nether_wart"},
    {
        "type": 116,
        "meta": 0,
        "name": "Enchantment Table",
        "text_type": "enchanting_table",
    },
    {"type": 117, "meta": 0, "name": "Brewing Stand", "text_type": "brewing_stand"},
    {"type": 118, "meta": 0, "name": "Cauldron", "text_type": "cauldron"},
    {"type": 119, "meta": 0, "name": "End Portal", "text_type": "end_portal"},
    {
        "type": 120,
        "meta": 0,
        "name": "End Portal Frame",
        "text_type": "end_portal_frame",
    },
    {"type": 121, "meta": 0, "name": "End Stone", "text_type": "end_stone"},
    {"type": 122, "meta": 0, "name": "Dragon Egg", "text_type": "dragon_egg"},
    {
        "type": 123,
        "meta": 0,
        "name": "Redstone Lamp (inactive)",
        "text_type": "redstone_lamp",
    },
    {
        "type": 124,
        "meta": 0,
        "name": "Redstone Lamp (active)",
        "text_type": "lit_redstone_lamp",
    },
    {
        "type": 125,
        "meta": 0,
        "name": "Double Oak Wood Slab",
        "text_type": "double_wooden_slab",
    },
    {
        "type": 125,
        "meta": 1,
        "name": "Double Spruce Wood Slab",
        "text_type": "double_wooden_slab",
    },
    {
        "type": 125,
        "meta": 2,
        "name": "Double Birch Wood Slab",
        "text_type": "double_wooden_slab",
    },
    {
        "type": 125,
        "meta": 3,
        "name": "Double Jungle Wood Slab",
        "text_type": "double_wooden_slab",
    },
    {
        "type": 125,
        "meta": 4,
        "name": "Double Acacia Wood Slab",
        "text_type": "double_wooden_slab",
    },
    {
        "type": 125,
        "meta": 5,
        "name": "Double Dark Oak Wood Slab",
        "text_type": "double_wooden_slab",
    },
    {"type": 126, "meta": 0, "name": "Oak Wood Slab", "text_type": "wooden_slab"},
    {"type": 126, "meta": 1, "name": "Spruce Wood Slab", "text_type": "wooden_slab"},
    {"type": 126, "meta": 2, "name": "Birch Wood Slab", "text_type": "wooden_slab"},
    {"type": 126, "meta": 3, "name": "Jungle Wood Slab", "text_type": "wooden_slab"},
    {"type": 126, "meta": 4, "name": "Acacia Wood Slab", "text_type": "wooden_slab"},
    {"type": 126, "meta": 5, "name": "Dark Oak Wood Slab", "text_type": "wooden_slab"},
    {"type": 127, "meta": 0, "name": "Cocoa", "text_type": "cocoa"},
    {
        "type": 128,
        "meta": 0,
        "name": "Sandstone Stairs",
        "text_type": "sandstone_stairs",
    },
    {"type": 129, "meta": 0, "name": "Emerald Ore", "text_type": "emerald_ore"},
    {"type": 130, "meta": 0, "name": "Ender Chest", "text_type": "ender_chest"},
    {"type": 131, "meta": 0, "name": "Tripwire Hook", "text_type": "tripwire_hook"},
    {"type": 132, "meta": 0, "name": "Tripwire", "text_type": "tripwire_hook"},
    {"type": 133, "meta": 0, "name": "Emerald Block", "text_type": "emerald_block"},
    {
        "type": 134,
        "meta": 0,
        "name": "Spruce Wood Stairs",
        "text_type": "spruce_stairs",
    },
    {"type": 135, "meta": 0, "name": "Birch Wood Stairs", "text_type": "birch_stairs"},
    {
        "type": 136,
        "meta": 0,
        "name": "Jungle Wood Stairs",
        "text_type": "jungle_stairs",
    },
    {"type": 137, "meta": 0, "name": "Command Block", "text_type": "command_block"},
    {"type": 138, "meta": 0, "name": "Beacon", "text_type": "beacon"},
    {
        "type": 139,
        "meta": 0,
        "name": "Cobblestone Wall",
        "text_type": "cobblestone_wall",
    },
    {
        "type": 139,
        "meta": 1,
        "name": "Mossy Cobblestone Wall",
        "text_type": "cobblestone_wall",
    },
    {"type": 140, "meta": 0, "name": "Flower Pot", "text_type": "flower_pot"},
    {"type": 141, "meta": 0, "name": "Carrots", "text_type": "carrots"},
    {"type": 142, "meta": 0, "name": "Potatoes", "text_type": "potatoes"},
    {"type": 143, "meta": 0, "name": "Wooden Button", "text_type": "wooden_button"},
    {"type": 144, "meta": 0, "name": "Mob Head", "text_type": "skull"},
    {"type": 145, "meta": 0, "name": "Anvil", "text_type": "anvil"},
    {"type": 146, "meta": 0, "name": "Trapped Chest", "text_type": "trapped_chest"},
    {
        "type": 147,
        "meta": 0,
        "name": "Weighted Pressure Plate (light)",
        "text_type": "light_weighted_pressure_plate",
    },
    {
        "type": 148,
        "meta": 0,
        "name": "Weighted Pressure Plate (heavy)",
        "text_type": "heavy_weighted_pressure_plate",
    },
    {
        "type": 149,
        "meta": 0,
        "name": "Redstone Comparator (inactive)",
        "text_type": "unpowered_comparator",
    },
    {
        "type": 150,
        "meta": 0,
        "name": "Redstone Comparator (active)",
        "text_type": "powered_comparator",
    },
    {
        "type": 151,
        "meta": 0,
        "name": "Daylight Sensor",
        "text_type": "daylight_detector",
    },
    {"type": 152, "meta": 0, "name": "Redstone Block", "text_type": "redstone_block"},
    {"type": 153, "meta": 0, "name": "Nether Quartz Ore", "text_type": "quartz_ore"},
    {"type": 154, "meta": 0, "name": "Hopper", "text_type": "hopper"},
    {"type": 155, "meta": 0, "name": "Quartz Block", "text_type": "quartz_block"},
    {
        "type": 155,
        "meta": 1,
        "name": "Chiseled Quartz Block",
        "text_type": "quartz_block",
    },
    {
        "type": 155,
        "meta": 2,
        "name": "Pillar Quartz Block",
        "text_type": "quartz_block",
    },
    {"type": 156, "meta": 0, "name": "Quartz Stairs", "text_type": "quartz_stairs"},
    {"type": 157, "meta": 0, "name": "Activator Rail", "text_type": "activator_rail"},
    {"type": 158, "meta": 0, "name": "Dropper", "text_type": "dropper"},
    {
        "type": 159,
        "meta": 0,
        "name": "White Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 1,
        "name": "Orange Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 2,
        "name": "Magenta Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 3,
        "name": "Light Blue Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 4,
        "name": "Yellow Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 5,
        "name": "Lime Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 6,
        "name": "Pink Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 7,
        "name": "Gray Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 8,
        "name": "Light Gray Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 9,
        "name": "Cyan Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 10,
        "name": "Purple Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 11,
        "name": "Blue Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 12,
        "name": "Brown Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 13,
        "name": "Green Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 14,
        "name": "Red Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 159,
        "meta": 15,
        "name": "Black Hardened Clay",
        "text_type": "stained_hardened_clay",
    },
    {
        "type": 160,
        "meta": 0,
        "name": "White Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 1,
        "name": "Orange Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 2,
        "name": "Magenta Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 3,
        "name": "Light Blue Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 4,
        "name": "Yellow Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 5,
        "name": "Lime Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 6,
        "name": "Pink Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 7,
        "name": "Gray Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 8,
        "name": "Light Gray Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 9,
        "name": "Cyan Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 10,
        "name": "Purple Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 11,
        "name": "Blue Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 12,
        "name": "Brown Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 13,
        "name": "Green Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 14,
        "name": "Red Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {
        "type": 160,
        "meta": 15,
        "name": "Black Stained Glass Pane",
        "text_type": "stained_glass_pane",
    },
    {"type": 161, "meta": 0, "name": "Acacia Leaves", "text_type": "leaves2"},
    {"type": 161, "meta": 1, "name": "Dark Oak Leaves", "text_type": "leaves2"},
    {"type": 162, "meta": 0, "name": "Acacia Wood", "text_type": "log2"},
    {"type": 162, "meta": 1, "name": "Dark Oak Wood", "text_type": "log2"},
    {
        "type": 163,
        "meta": 0,
        "name": "Acacia Wood Stairs",
        "text_type": "acacia_stairs",
    },
    {
        "type": 164,
        "meta": 0,
        "name": "Dark Oak Wood Stairs",
        "text_type": "dark_oak_stairs",
    },
    {"type": 165, "meta": 0, "name": "Slime Block", "text_type": "slime"},
    {"type": 166, "meta": 0, "name": "Barrier", "text_type": "barrier"},
    {"type": 167, "meta": 0, "name": "Iron Trapdoor", "text_type": "iron_trapdoor"},
    {"type": 168, "meta": 0, "name": "Prismarine", "text_type": "prismarine"},
    {"type": 168, "meta": 1, "name": "Prismarine Bricks", "text_type": "prismarine"},
    {"type": 168, "meta": 2, "name": "Dark Prismarine", "text_type": "prismarine"},
    {"type": 169, "meta": 0, "name": "Sea Lantern", "text_type": "sea_lantern"},
    {"type": 170, "meta": 0, "name": "Hay Bale", "text_type": "hay_block"},
    {"type": 171, "meta": 0, "name": "White Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 1, "name": "Orange Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 2, "name": "Magenta Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 3, "name": "Light Blue Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 4, "name": "Yellow Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 5, "name": "Lime Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 6, "name": "Pink Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 7, "name": "Gray Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 8, "name": "Light Gray Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 9, "name": "Cyan Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 10, "name": "Purple Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 11, "name": "Blue Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 12, "name": "Brown Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 13, "name": "Green Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 14, "name": "Red Carpet", "text_type": "carpet"},
    {"type": 171, "meta": 15, "name": "Black Carpet", "text_type": "carpet"},
    {"type": 172, "meta": 0, "name": "Hardened Clay", "text_type": "hardened_clay"},
    {"type": 173, "meta": 0, "name": "Block of Coal", "text_type": "coal_block"},
    {"type": 174, "meta": 0, "name": "Packed Ice", "text_type": "packed_ice"},
    {"type": 175, "meta": 0, "name": "Sunflower", "text_type": "double_plant"},
    {"type": 175, "meta": 1, "name": "Lilac", "text_type": "double_plant"},
    {"type": 175, "meta": 2, "name": "Double Tallgrass", "text_type": "double_plant"},
    {"type": 175, "meta": 3, "name": "Large Fern", "text_type": "double_plant"},
    {"type": 175, "meta": 4, "name": "Rose Bush", "text_type": "double_plant"},
    {"type": 175, "meta": 5, "name": "Peony", "text_type": "double_plant"},
    {
        "type": 176,
        "meta": 0,
        "name": "Free-standing Banner",
        "text_type": "standing_banner",
    },
    {"type": 177, "meta": 0, "name": "Wall-mounted Banner", "text_type": "wall_banner"},
    {
        "type": 178,
        "meta": 0,
        "name": "Inverted Daylight Sensor",
        "text_type": "daylight_detector_inverted",
    },
    {"type": 179, "meta": 0, "name": "Red Sandstone", "text_type": "red_sandstone"},
    {
        "type": 179,
        "meta": 1,
        "name": "Chiseled Red Sandstone",
        "text_type": "red_sandstone",
    },
    {
        "type": 179,
        "meta": 2,
        "name": "Smooth Red Sandstone",
        "text_type": "red_sandstone",
    },
    {
        "type": 180,
        "meta": 0,
        "name": "Red Sandstone Stairs",
        "text_type": "red_sandstone_stairs",
    },
    {
        "type": 181,
        "meta": 0,
        "name": "Double Red Sandstone Slab",
        "text_type": "double_stone_slab2",
    },
    {"type": 182, "meta": 0, "name": "Red Sandstone Slab", "text_type": "stone_slab2"},
    {
        "type": 183,
        "meta": 0,
        "name": "Spruce Fence Gate",
        "text_type": "spruce_fence_gate",
    },
    {
        "type": 184,
        "meta": 0,
        "name": "Birch Fence Gate",
        "text_type": "birch_fence_gate",
    },
    {
        "type": 185,
        "meta": 0,
        "name": "Jungle Fence Gate",
        "text_type": "jungle_fence_gate",
    },
    {
        "type": 186,
        "meta": 0,
        "name": "Dark Oak Fence Gate",
        "text_type": "dark_oak_fence_gate",
    },
    {
        "type": 187,
        "meta": 0,
        "name": "Acacia Fence Gate",
        "text_type": "acacia_fence_gate",
    },
    {"type": 188, "meta": 0, "name": "Spruce Fence", "text_type": "spruce_fence"},
    {"type": 189, "meta": 0, "name": "Birch Fence", "text_type": "birch_fence"},
    {"type": 190, "meta": 0, "name": "Jungle Fence", "text_type": "jungle_fence"},
    {"type": 191, "meta": 0, "name": "Dark Oak Fence", "text_type": "dark_oak_fence"},
    {"type": 192, "meta": 0, "name": "Acacia Fence", "text_type": "acacia_fence"},
    {"type": 193, "meta": 0, "name": "Spruce Door Block", "text_type": "spruce_door"},
    {"type": 194, "meta": 0, "name": "Birch Door Block", "text_type": "birch_door"},
    {"type": 195, "meta": 0, "name": "Jungle Door Block", "text_type": "jungle_door"},
    {"type": 196, "meta": 0, "name": "Acacia Door Block", "text_type": "acacia_door"},
    {
        "type": 197,
        "meta": 0,
        "name": "Dark Oak Door Block",
        "text_type": "dark_oak_door",
    },
    {"type": 198, "meta": 0, "name": "End Rod", "text_type": "end_rod"},
    {"type": 199, "meta": 0, "name": "Chorus Plant", "text_type": "chorus_plant"},
    {"type": 200, "meta": 0, "name": "Chorus Flower", "text_type": "chorus_flower"},
    {"type": 201, "meta": 0, "name": "Purpur Block", "text_type": "purpur_block"},
    {"type": 202, "meta": 0, "name": "Purpur Pillar", "text_type": "purpur_pillar"},
    {"type": 203, "meta": 0, "name": "Purpur Stairs", "text_type": "purpur_stairs"},
    {
        "type": 204,
        "meta": 0,
        "name": "Purpur Double Slab",
        "text_type": "purpur_double_slab",
    },
    {"type": 205, "meta": 0, "name": "Purpur Slab", "text_type": "purpur_slab"},
    {"type": 206, "meta": 0, "name": "End Stone Bricks", "text_type": "end_bricks"},
    {"type": 207, "meta": 0, "name": "Beetroot Block", "text_type": "beetroots"},
    {"type": 208, "meta": 0, "name": "Grass Path", "text_type": "grass_path"},
    {"type": 209, "meta": 0, "name": "End Gateway", "text_type": "end_gateway"},
    {
        "type": 210,
        "meta": 0,
        "name": "Repeating Command Block",
        "text_type": "repeating_command_block",
    },
    {
        "type": 211,
        "meta": 0,
        "name": "Chain Command Block",
        "text_type": "chain_command_block",
    },
    {"type": 212, "meta": 0, "name": "Frosted Ice", "text_type": "frosted_ice"},
    {"type": 213, "meta": 0, "name": "Magma Block", "text_type": "magma"},
    {
        "type": 214,
        "meta": 0,
        "name": "Nether Wart Block",
        "text_type": "nether_wart_block",
    },
    {
        "type": 215,
        "meta": 0,
        "name": "Red Nether Brick",
        "text_type": "red_nether_brick",
    },
    {"type": 216, "meta": 0, "name": "Bone Block", "text_type": "bone_block"},
    {"type": 217, "meta": 0, "name": "Structure Void", "text_type": "structure_void"},
    {"type": 218, "meta": 0, "name": "Observer", "text_type": "observer"},
    {
        "type": 219,
        "meta": 0,
        "name": "White Shulker Box",
        "text_type": "white_shulker_box",
    },
    {
        "type": 220,
        "meta": 0,
        "name": "Orange Shulker Box",
        "text_type": "orange_shulker_box",
    },
    {
        "type": 221,
        "meta": 0,
        "name": "Magenta Shulker Box",
        "text_type": "magenta_shulker_box",
    },
    {
        "type": 222,
        "meta": 0,
        "name": "Light Blue Shulker Box",
        "text_type": "light_blue_shulker_box",
    },
    {
        "type": 223,
        "meta": 0,
        "name": "Yellow Shulker Box",
        "text_type": "yellow_shulker_box",
    },
    {
        "type": 224,
        "meta": 0,
        "name": "Lime Shulker Box",
        "text_type": "lime_shulker_box",
    },
    {
        "type": 225,
        "meta": 0,
        "name": "Pink Shulker Box",
        "text_type": "pink_shulker_box",
    },
    {
        "type": 226,
        "meta": 0,
        "name": "Gray Shulker Box",
        "text_type": "gray_shulker_box",
    },
    {
        "type": 227,
        "meta": 0,
        "name": "Light Gray Shulker Box",
        "text_type": "silver_shulker_box",
    },
    {
        "type": 228,
        "meta": 0,
        "name": "Cyan Shulker Box",
        "text_type": "cyan_shulker_box",
    },
    {
        "type": 229,
        "meta": 0,
        "name": "Purple Shulker Box",
        "text_type": "purple_shulker_box",
    },
    {
        "type": 230,
        "meta": 0,
        "name": "Blue Shulker Box",
        "text_type": "blue_shulker_box",
    },
    {
        "type": 231,
        "meta": 0,
        "name": "Brown Shulker Box",
        "text_type": "brown_shulker_box",
    },
    {
        "type": 232,
        "meta": 0,
        "name": "Green Shulker Box",
        "text_type": "green_shulker_box",
    },
    {"type": 233, "meta": 0, "name": "Red Shulker Box", "text_type": "red_shulker_box"},
    {
        "type": 234,
        "meta": 0,
        "name": "Black Shulker Box",
        "text_type": "black_shulker_box",
    },
    {
        "type": 235,
        "meta": 0,
        "name": "White Glazed Terracotta",
        "text_type": "white_glazed_terracotta",
    },
    {
        "type": 236,
        "meta": 0,
        "name": "Orange Glazed Terracotta",
        "text_type": "orange_glazed_terracotta",
    },
    {
        "type": 237,
        "meta": 0,
        "name": "Magenta Glazed Terracotta",
        "text_type": "magenta_glazed_terracotta",
    },
    {
        "type": 238,
        "meta": 0,
        "name": "Light Blue Glazed Terracotta",
        "text_type": "light_blue_glazed_terracotta",
    },
    {
        "type": 239,
        "meta": 0,
        "name": "Yellow Glazed Terracotta",
        "text_type": "yellow_glazed_terracotta",
    },
    {
        "type": 240,
        "meta": 0,
        "name": "Lime Glazed Terracotta",
        "text_type": "lime_glazed_terracotta",
    },
    {
        "type": 241,
        "meta": 0,
        "name": "Pink Glazed Terracotta",
        "text_type": "pink_glazed_terracotta",
    },
    {
        "type": 242,
        "meta": 0,
        "name": "Gray Glazed Terracotta",
        "text_type": "gray_glazed_terracotta",
    },
    {
        "type": 243,
        "meta": 0,
        "name": "Light Gray Glazed Terracotta",
        "text_type": "light_gray_glazed_terracotta",
    },
    {
        "type": 244,
        "meta": 0,
        "name": "Cyan Glazed Terracotta",
        "text_type": "cyan_glazed_terracotta",
    },
    {
        "type": 245,
        "meta": 0,
        "name": "Purple Glazed Terracotta",
        "text_type": "purple_glazed_terracotta",
    },
    {
        "type": 246,
        "meta": 0,
        "name": "Blue Glazed Terracotta",
        "text_type": "blue_glazed_terracotta",
    },
    {
        "type": 247,
        "meta": 0,
        "name": "Brown Glazed Terracotta",
        "text_type": "brown_glazed_terracotta",
    },
    {
        "type": 248,
        "meta": 0,
        "name": "Green Glazed Terracotta",
        "text_type": "green_glazed_terracotta",
    },
    {
        "type": 249,
        "meta": 0,
        "name": "Red Glazed Terracotta",
        "text_type": "red_glazed_terracotta",
    },
    {
        "type": 250,
        "meta": 0,
        "name": "Black Glazed Terracotta",
        "text_type": "black_glazed_terracotta",
    },
    {"type": 251, "meta": 0, "name": "White Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 1, "name": "Orange Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 2, "name": "Magenta Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 3, "name": "Light Blue Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 4, "name": "Yellow Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 5, "name": "Lime Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 6, "name": "Pink Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 7, "name": "Gray Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 8, "name": "Light Gray Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 9, "name": "Cyan Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 10, "name": "Purple Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 11, "name": "Blue Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 12, "name": "Brown Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 13, "name": "Green Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 14, "name": "Red Concrete", "text_type": "concrete"},
    {"type": 251, "meta": 15, "name": "Black Concrete", "text_type": "concrete"},
    {
        "type": 252,
        "meta": 0,
        "name": "White Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 1,
        "name": "Orange Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 2,
        "name": "Magenta Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 3,
        "name": "Light Blue Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 4,
        "name": "Yellow Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 5,
        "name": "Lime Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 6,
        "name": "Pink Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 7,
        "name": "Gray Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 8,
        "name": "Light Gray Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 9,
        "name": "Cyan Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 10,
        "name": "Purple Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 11,
        "name": "Blue Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 12,
        "name": "Brown Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 13,
        "name": "Green Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 14,
        "name": "Red Concrete Powder",
        "text_type": "concrete_powder",
    },
    {
        "type": 252,
        "meta": 15,
        "name": "Black Concrete Powder",
        "text_type": "concrete_powder",
    },
    {"type": 255, "meta": 0, "name": "Structure Block", "text_type": "structure_block"},
    {"type": 256, "meta": 0, "name": "Iron Shovel", "text_type": "iron_shovel"},
    {"type": 257, "meta": 0, "name": "Iron Pickaxe", "text_type": "iron_pickaxe"},
    {"type": 258, "meta": 0, "name": "Iron Axe", "text_type": "iron_axe"},
    {"type": 259, "meta": 0, "name": "Flint and Steel", "text_type": "flint_and_steel"},
    {"type": 260, "meta": 0, "name": "Apple", "text_type": "apple"},
    {"type": 261, "meta": 0, "name": "Bow", "text_type": "bow"},
    {"type": 262, "meta": 0, "name": "Arrow", "text_type": "arrow"},
    {"type": 263, "meta": 0, "name": "Coal", "text_type": "coal"},
    {"type": 263, "meta": 1, "name": "Charcoal", "text_type": "coal"},
    {"type": 264, "meta": 0, "name": "Diamond", "text_type": "diamond"},
    {"type": 265, "meta": 0, "name": "Iron Ingot", "text_type": "iron_ingot"},
    {"type": 266, "meta": 0, "name": "Gold Ingot", "text_type": "gold_ingot"},
    {"type": 267, "meta": 0, "name": "Iron Sword", "text_type": "iron_sword"},
    {"type": 268, "meta": 0, "name": "Wooden Sword", "text_type": "wooden_sword"},
    {"type": 269, "meta": 0, "name": "Wooden Shovel", "text_type": "wooden_shovel"},
    {"type": 270, "meta": 0, "name": "Wooden Pickaxe", "text_type": "wooden_pickaxe"},
    {"type": 271, "meta": 0, "name": "Wooden Axe", "text_type": "wooden_axe"},
    {"type": 272, "meta": 0, "name": "Stone Sword", "text_type": "stone_sword"},
    {"type": 273, "meta": 0, "name": "Stone Shovel", "text_type": "stone_shovel"},
    {"type": 274, "meta": 0, "name": "Stone Pickaxe", "text_type": "stone_pickaxe"},
    {"type": 275, "meta": 0, "name": "Stone Axe", "text_type": "stone_axe"},
    {"type": 276, "meta": 0, "name": "Diamond Sword", "text_type": "diamond_sword"},
    {"type": 277, "meta": 0, "name": "Diamond Shovel", "text_type": "diamond_shovel"},
    {"type": 278, "meta": 0, "name": "Diamond Pickaxe", "text_type": "diamond_pickaxe"},
    {"type": 279, "meta": 0, "name": "Diamond Axe", "text_type": "diamond_axe"},
    {"type": 280, "meta": 0, "name": "Stick", "text_type": "stick"},
    {"type": 281, "meta": 0, "name": "Bowl", "text_type": "bowl"},
    {"type": 282, "meta": 0, "name": "Mushroom Stew", "text_type": "mushroom_stew"},
    {"type": 283, "meta": 0, "name": "Golden Sword", "text_type": "golden_sword"},
    {"type": 284, "meta": 0, "name": "Golden Shovel", "text_type": "golden_shovel"},
    {"type": 285, "meta": 0, "name": "Golden Pickaxe", "text_type": "golden_pickaxe"},
    {"type": 286, "meta": 0, "name": "Golden Axe", "text_type": "golden_axe"},
    {"type": 287, "meta": 0, "name": "String", "text_type": "string"},
    {"type": 288, "meta": 0, "name": "Feather", "text_type": "feather"},
    {"type": 289, "meta": 0, "name": "Gunpowder", "text_type": "gunpowder"},
    {"type": 290, "meta": 0, "name": "Wooden Hoe", "text_type": "wooden_hoe"},
    {"type": 291, "meta": 0, "name": "Stone Hoe", "text_type": "stone_hoe"},
    {"type": 292, "meta": 0, "name": "Iron Hoe", "text_type": "iron_hoe"},
    {"type": 293, "meta": 0, "name": "Diamond Hoe", "text_type": "diamond_hoe"},
    {"type": 294, "meta": 0, "name": "Golden Hoe", "text_type": "golden_hoe"},
    {"type": 295, "meta": 0, "name": "Wheat Seeds", "text_type": "wheat_seeds"},
    {"type": 296, "meta": 0, "name": "Wheat", "text_type": "wheat"},
    {"type": 297, "meta": 0, "name": "Bread", "text_type": "bread"},
    {"type": 298, "meta": 0, "name": "Leather Helmet", "text_type": "leather_helmet"},
    {
        "type": 299,
        "meta": 0,
        "name": "Leather Tunic",
        "text_type": "leather_chestplate",
    },
    {"type": 300, "meta": 0, "name": "Leather Pants", "text_type": "leather_leggings"},
    {"type": 301, "meta": 0, "name": "Leather Boots", "text_type": "leather_boots"},
    {
        "type": 302,
        "meta": 0,
        "name": "Chainmail Helmet",
        "text_type": "chainmail_helmet",
    },
    {
        "type": 303,
        "meta": 0,
        "name": "Chainmail Chestplate",
        "text_type": "chainmail_chestplate",
    },
    {
        "type": 304,
        "meta": 0,
        "name": "Chainmail Leggings",
        "text_type": "chainmail_leggings",
    },
    {"type": 305, "meta": 0, "name": "Chainmail Boots", "text_type": "chainmail_boots"},
    {"type": 306, "meta": 0, "name": "Iron Helmet", "text_type": "iron_helmet"},
    {"type": 307, "meta": 0, "name": "Iron Chestplate", "text_type": "iron_chestplate"},
    {"type": 308, "meta": 0, "name": "Iron Leggings", "text_type": "iron_leggings"},
    {"type": 309, "meta": 0, "name": "Iron Boots", "text_type": "iron_boots"},
    {"type": 310, "meta": 0, "name": "Diamond Helmet", "text_type": "diamond_helmet"},
    {
        "type": 311,
        "meta": 0,
        "name": "Diamond Chestplate",
        "text_type": "diamond_chestplate",
    },
    {
        "type": 312,
        "meta": 0,
        "name": "Diamond Leggings",
        "text_type": "diamond_leggings",
    },
    {"type": 313, "meta": 0, "name": "Diamond Boots", "text_type": "diamond_boots"},
    {"type": 314, "meta": 0, "name": "Golden Helmet", "text_type": "golden_helmet"},
    {
        "type": 315,
        "meta": 0,
        "name": "Golden Chestplate",
        "text_type": "golden_chestplate",
    },
    {"type": 316, "meta": 0, "name": "Golden Leggings", "text_type": "golden_leggings"},
    {"type": 317, "meta": 0, "name": "Golden Boots", "text_type": "golden_boots"},
    {"type": 318, "meta": 0, "name": "Flint", "text_type": "flint"},
    {"type": 319, "meta": 0, "name": "Raw Porkchop", "text_type": "porkchop"},
    {"type": 320, "meta": 0, "name": "Cooked Porkchop", "text_type": "cooked_porkchop"},
    {"type": 321, "meta": 0, "name": "Painting", "text_type": "painting"},
    {"type": 322, "meta": 0, "name": "Golden Apple", "text_type": "golden_apple"},
    {
        "type": 322,
        "meta": 1,
        "name": "Enchanted Golden Apple",
        "text_type": "golden_apple",
    },
    {"type": 323, "meta": 0, "name": "Sign", "text_type": "sign"},
    {"type": 324, "meta": 0, "name": "Oak Door", "text_type": "wooden_door"},
    {"type": 325, "meta": 0, "name": "Bucket", "text_type": "bucket"},
    {"type": 326, "meta": 0, "name": "Water Bucket", "text_type": "water_bucket"},
    {"type": 327, "meta": 0, "name": "Lava Bucket", "text_type": "lava_bucket"},
    {"type": 328, "meta": 0, "name": "Minecart", "text_type": "minecart"},
    {"type": 329, "meta": 0, "name": "Saddle", "text_type": "saddle"},
    {"type": 330, "meta": 0, "name": "Iron Door", "text_type": "iron_door"},
    {"type": 331, "meta": 0, "name": "Redstone", "text_type": "redstone"},
    {"type": 332, "meta": 0, "name": "Snowball", "text_type": "snowball"},
    {"type": 333, "meta": 0, "name": "Oak Boat", "text_type": "boat"},
    {"type": 334, "meta": 0, "name": "Leather", "text_type": "leather"},
    {"type": 335, "meta": 0, "name": "Milk Bucket", "text_type": "milk_bucket"},
    {"type": 336, "meta": 0, "name": "Brick", "text_type": "brick"},
    {"type": 337, "meta": 0, "name": "Clay", "text_type": "clay_ball"},
    {"type": 338, "meta": 0, "name": "Sugar Canes", "text_type": "reeds"},
    {"type": 339, "meta": 0, "name": "Paper", "text_type": "paper"},
    {"type": 340, "meta": 0, "name": "Book", "text_type": "book"},
    {"type": 341, "meta": 0, "name": "Slimeball", "text_type": "slime_ball"},
    {
        "type": 342,
        "meta": 0,
        "name": "Minecart with Chest",
        "text_type": "chest_minecart",
    },
    {
        "type": 343,
        "meta": 0,
        "name": "Minecart with Furnace",
        "text_type": "furnace_minecart",
    },
    {"type": 344, "meta": 0, "name": "Egg", "text_type": "egg"},
    {"type": 345, "meta": 0, "name": "Compass", "text_type": "compass"},
    {"type": 346, "meta": 0, "name": "Fishing Rod", "text_type": "fishing_rod"},
    {"type": 347, "meta": 0, "name": "Clock", "text_type": "clock"},
    {"type": 348, "meta": 0, "name": "Glowstone Dust", "text_type": "glowstone_dust"},
    {"type": 349, "meta": 0, "name": "Raw Fish", "text_type": "fish"},
    {"type": 349, "meta": 1, "name": "Raw Salmon", "text_type": "fish"},
    {"type": 349, "meta": 2, "name": "Clownfish", "text_type": "fish"},
    {"type": 349, "meta": 3, "name": "Pufferfish", "text_type": "fish"},
    {"type": 350, "meta": 0, "name": "Cooked Fish", "text_type": "cooked_fish"},
    {"type": 350, "meta": 1, "name": "Cooked Salmon", "text_type": "cooked_fish"},
    {"type": 351, "meta": 0, "name": "Ink Sack", "text_type": "dye"},
    {"type": 351, "meta": 1, "name": "Rose Red", "text_type": "dye"},
    {"type": 351, "meta": 2, "name": "Cactus Green", "text_type": "dye"},
    {"type": 351, "meta": 3, "name": "Coco Beans", "text_type": "dye"},
    {"type": 351, "meta": 4, "name": "Lapis Lazuli", "text_type": "dye"},
    {"type": 351, "meta": 5, "name": "Purple Dye", "text_type": "dye"},
    {"type": 351, "meta": 6, "name": "Cyan Dye", "text_type": "dye"},
    {"type": 351, "meta": 7, "name": "Light Gray Dye", "text_type": "dye"},
    {"type": 351, "meta": 8, "name": "Gray Dye", "text_type": "dye"},
    {"type": 351, "meta": 9, "name": "Pink Dye", "text_type": "dye"},
    {"type": 351, "meta": 10, "name": "Lime Dye", "text_type": "dye"},
    {"type": 351, "meta": 11, "name": "Dandelion Yellow", "text_type": "dye"},
    {"type": 351, "meta": 12, "name": "Light Blue Dye", "text_type": "dye"},
    {"type": 351, "meta": 13, "name": "Magenta Dye", "text_type": "dye"},
    {"type": 351, "meta": 14, "name": "Orange Dye", "text_type": "dye"},
    {"type": 351, "meta": 15, "name": "Bone Meal", "text_type": "dye"},
    {"type": 352, "meta": 0, "name": "Bone", "text_type": "bone"},
    {"type": 353, "meta": 0, "name": "Sugar", "text_type": "sugar"},
    {"type": 354, "meta": 0, "name": "Cake", "text_type": "cake"},
    {"type": 355, "meta": 0, "name": "Bed", "text_type": "bed"},
    {"type": 356, "meta": 0, "name": "Redstone Repeater", "text_type": "repeater"},
    {"type": 357, "meta": 0, "name": "Cookie", "text_type": "cookie"},
    {"type": 358, "meta": 0, "name": "Map", "text_type": "filled_map"},
    {"type": 359, "meta": 0, "name": "Shears", "text_type": "shears"},
    {"type": 360, "meta": 0, "name": "Melon", "text_type": "melon"},
    {"type": 361, "meta": 0, "name": "Pumpkin Seeds", "text_type": "pumpkin_seeds"},
    {"type": 362, "meta": 0, "name": "Melon Seeds", "text_type": "melon_seeds"},
    {"type": 363, "meta": 0, "name": "Raw Beef", "text_type": "beef"},
    {"type": 364, "meta": 0, "name": "Steak", "text_type": "cooked_beef"},
    {"type": 365, "meta": 0, "name": "Raw Chicken", "text_type": "chicken"},
    {"type": 366, "meta": 0, "name": "Cooked Chicken", "text_type": "cooked_chicken"},
    {"type": 367, "meta": 0, "name": "Rotten Flesh", "text_type": "rotten_flesh"},
    {"type": 368, "meta": 0, "name": "Ender Pearl", "text_type": "ender_pearl"},
    {"type": 369, "meta": 0, "name": "Blaze Rod", "text_type": "blaze_rod"},
    {"type": 370, "meta": 0, "name": "Ghast Tear", "text_type": "ghast_tear"},
    {"type": 371, "meta": 0, "name": "Gold Nugget", "text_type": "gold_nugget"},
    {"type": 372, "meta": 0, "name": "Nether Wart", "text_type": "nether_wart"},
    {"type": 373, "meta": 0, "name": "Potion", "text_type": "potion"},
    {"type": 374, "meta": 0, "name": "Glass Bottle", "text_type": "glass_bottle"},
    {"type": 375, "meta": 0, "name": "Spider Eye", "text_type": "spider_eye"},
    {
        "type": 376,
        "meta": 0,
        "name": "Fermented Spider Eye",
        "text_type": "fermented_spider_eye",
    },
    {"type": 377, "meta": 0, "name": "Blaze Powder", "text_type": "blaze_powder"},
    {"type": 378, "meta": 0, "name": "Magma Cream", "text_type": "magma_cream"},
    {"type": 379, "meta": 0, "name": "Brewing Stand", "text_type": "brewing_stand"},
    {"type": 380, "meta": 0, "name": "Cauldron", "text_type": "cauldron"},
    {"type": 381, "meta": 0, "name": "Eye of Ender", "text_type": "ender_eye"},
    {"type": 382, "meta": 0, "name": "Glistering Melon", "text_type": "speckled_melon"},
    {"type": 383, "meta": 4, "name": "Spawn Elder Guardian", "text_type": "spawn_egg"},
    {"type": 383, "meta": 5, "name": "Spawn Wither Skeleton", "text_type": "spawn_egg"},
    {"type": 383, "meta": 6, "name": "Spawn Stray", "text_type": "spawn_egg"},
    {"type": 383, "meta": 23, "name": "Spawn Husk", "text_type": "spawn_egg"},
    {
        "type": 383,
        "meta": 27,
        "name": "Spawn Zombie Villager",
        "text_type": "spawn_egg",
    },
    {"type": 383, "meta": 28, "name": "Spawn Skeleton Horse", "text_type": "spawn_egg"},
    {"type": 383, "meta": 29, "name": "Spawn Zombie Horse", "text_type": "spawn_egg"},
    {"type": 383, "meta": 31, "name": "Spawn Donkey", "text_type": "spawn_egg"},
    {"type": 383, "meta": 32, "name": "Spawn Mule", "text_type": "spawn_egg"},
    {"type": 383, "meta": 34, "name": "Spawn Evoker", "text_type": "spawn_egg"},
    {"type": 383, "meta": 35, "name": "Spawn Vex", "text_type": "spawn_egg"},
    {"type": 383, "meta": 36, "name": "Spawn Vindicator", "text_type": "spawn_egg"},
    {"type": 383, "meta": 50, "name": "Spawn Creeper", "text_type": "spawn_egg"},
    {"type": 383, "meta": 51, "name": "Spawn Skeleton", "text_type": "spawn_egg"},
    {"type": 383, "meta": 52, "name": "Spawn Spider", "text_type": "spawn_egg"},
    {"type": 383, "meta": 54, "name": "Spawn Zombie", "text_type": "spawn_egg"},
    {"type": 383, "meta": 55, "name": "Spawn Slime", "text_type": "spawn_egg"},
    {"type": 383, "meta": 56, "name": "Spawn Ghast", "text_type": "spawn_egg"},
    {"type": 383, "meta": 57, "name": "Spawn Zombie Pigman", "text_type": "spawn_egg"},
    {"type": 383, "meta": 58, "name": "Spawn Enderman", "text_type": "spawn_egg"},
    {"type": 383, "meta": 59, "name": "Spawn Cave Spider", "text_type": "spawn_egg"},
    {"type": 383, "meta": 60, "name": "Spawn Silverfish", "text_type": "spawn_egg"},
    {"type": 383, "meta": 61, "name": "Spawn Blaze", "text_type": "spawn_egg"},
    {"type": 383, "meta": 62, "name": "Spawn Magma Cube", "text_type": "spawn_egg"},
    {"type": 383, "meta": 65, "name": "Spawn Bat", "text_type": "spawn_egg"},
    {"type": 383, "meta": 66, "name": "Spawn Witch", "text_type": "spawn_egg"},
    {"type": 383, "meta": 67, "name": "Spawn Endermite", "text_type": "spawn_egg"},
    {"type": 383, "meta": 68, "name": "Spawn Guardian", "text_type": "spawn_egg"},
    {"type": 383, "meta": 69, "name": "Spawn Shulker", "text_type": "spawn_egg"},
    {"type": 383, "meta": 90, "name": "Spawn Pig", "text_type": "spawn_egg"},
    {"type": 383, "meta": 91, "name": "Spawn Sheep", "text_type": "spawn_egg"},
    {"type": 383, "meta": 92, "name": "Spawn Cow", "text_type": "spawn_egg"},
    {"type": 383, "meta": 93, "name": "Spawn Chicken", "text_type": "spawn_egg"},
    {"type": 383, "meta": 94, "name": "Spawn Squid", "text_type": "spawn_egg"},
    {"type": 383, "meta": 95, "name": "Spawn Wolf", "text_type": "spawn_egg"},
    {"type": 383, "meta": 96, "name": "Spawn Mooshroom", "text_type": "spawn_egg"},
    {"type": 383, "meta": 98, "name": "Spawn Ocelot", "text_type": "spawn_egg"},
    {"type": 383, "meta": 100, "name": "Spawn Horse", "text_type": "spawn_egg"},
    {"type": 383, "meta": 101, "name": "Spawn Rabbit", "text_type": "spawn_egg"},
    {"type": 383, "meta": 102, "name": "Spawn Polar Bear", "text_type": "spawn_egg"},
    {"type": 383, "meta": 103, "name": "Spawn Llama", "text_type": "spawn_egg"},
    {"type": 383, "meta": 105, "name": "Spawn Parrot", "text_type": "spawn_egg"},
    {"type": 383, "meta": 120, "name": "Spawn Villager", "text_type": "spawn_egg"},
    {
        "type": 384,
        "meta": 0,
        "name": "Bottle o' Enchanting",
        "text_type": "experience_bottle",
    },
    {"type": 385, "meta": 0, "name": "Fire Charge", "text_type": "fire_charge"},
    {"type": 386, "meta": 0, "name": "Book and Quill", "text_type": "writable_book"},
    {"type": 387, "meta": 0, "name": "Written Book", "text_type": "written_book"},
    {"type": 388, "meta": 0, "name": "Emerald", "text_type": "emerald"},
    {"type": 389, "meta": 0, "name": "Item Frame", "text_type": "item_frame"},
    {"type": 390, "meta": 0, "name": "Flower Pot", "text_type": "flower_pot"},
    {"type": 391, "meta": 0, "name": "Carrot", "text_type": "carrot"},
    {"type": 392, "meta": 0, "name": "Potato", "text_type": "potato"},
    {"type": 393, "meta": 0, "name": "Baked Potato", "text_type": "baked_potato"},
    {
        "type": 394,
        "meta": 0,
        "name": "Poisonous Potato",
        "text_type": "poisonous_potato",
    },
    {"type": 395, "meta": 0, "name": "Empty Map", "text_type": "map"},
    {"type": 396, "meta": 0, "name": "Golden Carrot", "text_type": "golden_carrot"},
    {"type": 397, "meta": 0, "name": "Mob Head (Skeleton)", "text_type": "skull"},
    {
        "type": 397,
        "meta": 1,
        "name": "Mob Head (Wither Skeleton)",
        "text_type": "skull",
    },
    {"type": 397, "meta": 2, "name": "Mob Head (Zombie)", "text_type": "skull"},
    {"type": 397, "meta": 3, "name": "Mob Head (Human)", "text_type": "skull"},
    {"type": 397, "meta": 4, "name": "Mob Head (Creeper)", "text_type": "skull"},
    {"type": 397, "meta": 5, "name": "Mob Head (Dragon)", "text_type": "skull"},
    {
        "type": 398,
        "meta": 0,
        "name": "Carrot on a Stick",
        "text_type": "carrot_on_a_stick",
    },
    {"type": 399, "meta": 0, "name": "Nether Star", "text_type": "nether_star"},
    {"type": 400, "meta": 0, "name": "Pumpkin Pie", "text_type": "pumpkin_pie"},
    {"type": 401, "meta": 0, "name": "Firework Rocket", "text_type": "fireworks"},
    {"type": 402, "meta": 0, "name": "Firework Star", "text_type": "firework_charge"},
    {"type": 403, "meta": 0, "name": "Enchanted Book", "text_type": "enchanted_book"},
    {"type": 404, "meta": 0, "name": "Redstone Comparator", "text_type": "comparator"},
    {"type": 405, "meta": 0, "name": "Nether Brick", "text_type": "netherbrick"},
    {"type": 406, "meta": 0, "name": "Nether Quartz", "text_type": "quartz"},
    {"type": 407, "meta": 0, "name": "Minecart with TNT", "text_type": "tnt_minecart"},
    {
        "type": 408,
        "meta": 0,
        "name": "Minecart with Hopper",
        "text_type": "hopper_minecart",
    },
    {
        "type": 409,
        "meta": 0,
        "name": "Prismarine Shard",
        "text_type": "prismarine_shard",
    },
    {
        "type": 410,
        "meta": 0,
        "name": "Prismarine Crystals",
        "text_type": "prismarine_crystals",
    },
    {"type": 411, "meta": 0, "name": "Raw Rabbit", "text_type": "rabbit"},
    {"type": 412, "meta": 0, "name": "Cooked Rabbit", "text_type": "cooked_rabbit"},
    {"type": 413, "meta": 0, "name": "Rabbit Stew", "text_type": "rabbit_stew"},
    {"type": 414, "meta": 0, "name": "Rabbit's Foot", "text_type": "rabbit_foot"},
    {"type": 415, "meta": 0, "name": "Rabbit Hide", "text_type": "rabbit_hide"},
    {"type": 416, "meta": 0, "name": "Armor Stand", "text_type": "armor_stand"},
    {
        "type": 417,
        "meta": 0,
        "name": "Iron Horse Armor",
        "text_type": "iron_horse_armor",
    },
    {
        "type": 418,
        "meta": 0,
        "name": "Golden Horse Armor",
        "text_type": "golden_horse_armor",
    },
    {
        "type": 419,
        "meta": 0,
        "name": "Diamond Horse Armor",
        "text_type": "diamond_horse_armor",
    },
    {"type": 420, "meta": 0, "name": "Lead", "text_type": "lead"},
    {"type": 421, "meta": 0, "name": "Name Tag", "text_type": "name_tag"},
    {
        "type": 422,
        "meta": 0,
        "name": "Minecart with Command Block",
        "text_type": "command_block_minecart",
    },
    {"type": 423, "meta": 0, "name": "Raw Mutton", "text_type": "mutton"},
    {"type": 424, "meta": 0, "name": "Cooked Mutton", "text_type": "cooked_mutton"},
    {"type": 425, "meta": 0, "name": "Banner", "text_type": "banner"},
    {"type": 426, "meta": 0, "name": "End Crystal", "text_type": "end_crystal"},
    {"type": 427, "meta": 0, "name": "Spruce Door", "text_type": "spruce_door"},
    {"type": 428, "meta": 0, "name": "Birch Door", "text_type": "birch_door"},
    {"type": 429, "meta": 0, "name": "Jungle Door", "text_type": "jungle_door"},
    {"type": 430, "meta": 0, "name": "Acacia Door", "text_type": "acacia_door"},
    {"type": 431, "meta": 0, "name": "Dark Oak Door", "text_type": "dark_oak_door"},
    {"type": 432, "meta": 0, "name": "Chorus Fruit", "text_type": "chorus_fruit"},
    {
        "type": 433,
        "meta": 0,
        "name": "Popped Chorus Fruit",
        "text_type": "popped_chorus_fruit",
    },
    {"type": 434, "meta": 0, "name": "Beetroot", "text_type": "beetroot"},
    {"type": 435, "meta": 0, "name": "Beetroot Seeds", "text_type": "beetroot_seeds"},
    {"type": 436, "meta": 0, "name": "Beetroot Soup", "text_type": "beetroot_soup"},
    {"type": 437, "meta": 0, "name": "Dragon's Breath", "text_type": "dragon_breath"},
    {"type": 438, "meta": 0, "name": "Splash Potion", "text_type": "splash_potion"},
    {"type": 439, "meta": 0, "name": "Spectral Arrow", "text_type": "spectral_arrow"},
    {"type": 440, "meta": 0, "name": "Tipped Arrow", "text_type": "tipped_arrow"},
    {
        "type": 441,
        "meta": 0,
        "name": "Lingering Potion",
        "text_type": "lingering_potion",
    },
    {"type": 442, "meta": 0, "name": "Shield", "text_type": "shield"},
    {"type": 443, "meta": 0, "name": "Elytra", "text_type": "elytra"},
    {"type": 444, "meta": 0, "name": "Spruce Boat", "text_type": "spruce_boat"},
    {"type": 445, "meta": 0, "name": "Birch Boat", "text_type": "birch_boat"},
    {"type": 446, "meta": 0, "name": "Jungle Boat", "text_type": "jungle_boat"},
    {"type": 447, "meta": 0, "name": "Acacia Boat", "text_type": "acacia_boat"},
    {"type": 448, "meta": 0, "name": "Dark Oak Boat", "text_type": "dark_oak_boat"},
    {
        "type": 449,
        "meta": 0,
        "name": "Totem of Undying",
        "text_type": "totem_of_undying",
    },
    {"type": 450, "meta": 0, "name": "Shulker Shell", "text_type": "shulker_shell"},
    {"type": 452, "meta": 0, "name": "Iron Nugget", "text_type": "iron_nugget"},
    {"type": 453, "meta": 0, "name": "Knowledge Book", "text_type": "knowledge_book"},
    {"type": 2256, "meta": 0, "name": "13 Disc", "text_type": "record_13"},
    {"type": 2257, "meta": 0, "name": "Cat Disc", "text_type": "record_cat"},
    {"type": 2258, "meta": 0, "name": "Blocks Disc", "text_type": "record_blocks"},
    {"type": 2259, "meta": 0, "name": "Chirp Disc", "text_type": "record_chirp"},
    {"type": 2260, "meta": 0, "name": "Far Disc", "text_type": "record_far"},
    {"type": 2261, "meta": 0, "name": "Mall Disc", "text_type": "record_mall"},
    {"type": 2262, "meta": 0, "name": "Mellohi Disc", "text_type": "record_mellohi"},
    {"type": 2263, "meta": 0, "name": "Stal Disc", "text_type": "record_stal"},
    {"type": 2264, "meta": 0, "name": "Strad Disc", "text_type": "record_strad"},
    {"type": 2265, "meta": 0, "name": "Ward Disc", "text_type": "record_ward"},
    {"type": 2266, "meta": 0, "name": "11 Disc", "text_type": "record_11"},
    {"type": 2267, "meta": 0, "name": "Wait Disc", "text_type": "record_wait"},
]

scs = [
    "pond_squid",
    "sea_walker",
    "night_squid",
    "frozen_steve",
    "sea_guardian",
    "frosty_the_snowman",
    "scarecrow",
    "sea_witch",
    "sea_archer",
    "oasis_rabbit",
    "oasis_sheep",
    "zombie_deep",
    "chicken_deep",
    "grinch",
    "catfish",
    "nightmare",
    "carrot_king",
    "water_worm",
    "sea_leech",
    "poisoned_water_worm",
    "flaming_worm",
    "guardian_defender",
    "werewolf",
    "deep_sea_protector",
    "lava_blaze",
    "lava_pigman",
    "water_hydra",
    "guardian_emperor",
    "skeleton_emperor",
    "phantom_fisherman",
    "zombie_miner",
    "yeti",
    "grim_reaper",
    "nurse_shark",
    "blue_shark",
    "tiger_shark",
    "great_white_shark",
]



from .vars import cata_xp_cumulative
from .hypyobject import HypyObject


class SkyBlockResources(HypyObject):
    """Hypixel SkyBlock resources"""

    __slots__ = ("_raw", "cumulative_xp")

    def __init__(self, data) -> None:
        self._raw = data["skills"]
        self.update_cumulative_xp()

    def update_cumulative_xp(self) -> None:
        """Update Internal cumulative XP Tables"""
        cumu_xp: dict = {}
        for skill, values in self._raw.items():
            i_skill = str.lower(skill)
            cumu_xp[i_skill] = {}
            cumu_xp[i_skill]["details"] = {}
            cumu_xp[i_skill]["levels"] = {0: 0}
            cumu_xp[i_skill]["details"]["max_level"] = values["maxLevel"]
            for level in values["levels"]:
                cumu_xp[i_skill]["levels"][level["level"]] = int(
                    level["totalExpRequired"]
                )
        cumu_xp["catacombs"] = {}
        cumu_xp["catacombs"]["levels"] = cata_xp_cumulative
        cumu_xp["catacombs"]["details"] = {"max_level": 50}
        self.cumulative_xp = cumu_xp



import datetime as dt
import asyncio
from typing import Iterator, List
from .exceptions import NotInGuildException
from .uuid import UUID
from .player import Player
from .hypyobject import HypyObject
from .abc import HypyIterable
from .guildrank import GuildRank
from .guildmember import GuildMember


class Guild(HypyObject, HypyIterable):
    """A Hypixel Guild"""

    __slots__ = (
        "_raw",
        "_hypy",
        "_id",
    )

    name: str
    """The name of the guild"""
    members: List[GuildMember]
    """A list of GuildMembers"""

    def __init__(self, data: dict, hypy) -> None:
        self._hypy = hypy
        self._raw: dict = data["guild"]
        self.name = self._raw["name"]
        self.members: List[GuildMember] = [
            GuildMember(x, hypy) for x in self._raw["members"]
        ]
        self._id = self._raw["_id"]

    def __len__(self) -> int:
        return len(self.members)

    def __iter__(self) -> Iterator[GuildMember]:
        for i in self.members:
            yield i

    def iter_uuids(self) -> Iterator[UUID]:
        for i in self.members:
            yield UUID(i["uuid"])

    @property
    def uuid_list(self) -> List[str]:
        """List of UUIDs of players in guild"""
        li = []
        for mem in self._raw["members"]:
            li.append(UUID(mem["uuid"]).no_dashes)
        return li

    async def name_list(self) -> List[str]:
        """Get list of names of players in guild"""
        return await asyncio.gather(
            *[self._hypy.mojang.uuid_to_name(uuid) for uuid in self.uuid_list]
        )

    async def player_list(self) -> List[Player]:
        """Get list of hypy.Player objects for players in guild"""
        return await asyncio.gather(
            *[self._hypy.get_player(uuid) for uuid in self.uuid_list]
        )

    def when_joined(self, uuid) -> dict:
        """Check when a player joined the guild

        :param uuid: The uuid of the player
        """
        tr = {"inGuild": False, "joinedAt": dt.datetime.utcfromtimestamp(0)}
        uuid = UUID(uuid).no_dashes
        m = [x for x in self.members if x["uuid"] == uuid]
        if len(m) > 0:
            tr["inGuild"] = True
            tr["joinedAt"] = dt.datetime.utcfromtimestamp(m[0]["joined"] / 1000)
        return tr

    def rank_of(self, uuid) -> GuildRank:
        """Check rank of a player

        :param uuid: The uuid of the player
        """
        uuid = UUID(uuid).no_dashes
        m = [x for x in self.members if x.uuid == uuid]
        if len(m) != 0:
            mem = m[0]
            return [r for r in self.ranks if r.name == mem.rank][0]
        raise NotInGuildException(uuid)

    @property
    def ranks(self) -> List[GuildRank]:
        """The ranks of the guild"""
        tr = []
        i = 0
        for v in self._raw["ranks"]:
            tr.append(v)
            i += 1
        tr.sort(key=lambda x: x["priority"])
        rank_names = [x["name"] for x in tr]
        gm = [x["rank"] for x in self.members if x["rank"] not in rank_names][0]
        tr.append(
            {
                "name": gm,
                "default": False,
                "tag": "GM",
                "created": self._raw["created"],
                "priority": tr[-1]["priority"] + 1,
            }
        )
        return [GuildRank(x, idx) for idx, x in enumerate(tr)]



from typing import Dict
from collections import defaultdict
from orjson import orjson
import aiofiles
from .vars import scs
from .ext.senither.senither import SenitherWeight
from .hypyobject import HypyObject
from .skyblockskills import (
    SkyblockSkills,
    SkyblockSlayers,
    SkyblockBank,
    REGULAR_SKILLS,
    REGULAR_SKILL_FILTER,
)
from .skyblockinventories import SkyblockInventory, SkyblockBackpacks
from .exceptions import MaroNoSuccess


class SkyblockProfile(HypyObject):
    """A Hypixel SkyBlock Profile"""

    __slots__ = (
        "_raw",
        "_raw_member_info",
        "_hypy",
        "_uuid",
    )

    deleted: bool
    """Whether the profile has been deleted"""
    profile_id: str
    """The profile id of the profile"""

    def __init__(self, data, uuid, hypy) -> None:
        self._hypy = hypy
        self._raw = data["profile"]
        self.profile_id = self._raw["profile_id"]
        self._uuid = uuid
        self._raw_member_info = self._raw["members"][uuid]
        self.deleted = False
        if "last_save" not in self._raw_member_info:
            self._deleted = True
            return
        self.last_save = self._raw_member_info["last_save"]
        self.skills_api_on = any(
            [
                (REGULAR_SKILL_FILTER.format(x) in self._raw_member_info)
                for x in REGULAR_SKILLS
            ]
        )
        self.cute_name = None

    def add_cute_name(self, cute_name):
        """add a cute name to this profile"""
        self.cute_name = cute_name

    async def refetch(self) -> None:
        """Refresh the profile"""
        self._raw = (await self._hypy.get_profile_data(self.profile_id))["profile"]
        self._raw_member_info = self._raw["members"][self._uuid]

    async def save_json(self, filename) -> None:
        """Save raw member info to the disk

        :param filename: The name of the file to save to
        """
        async with aiofiles.open(filename, mode="w+") as f:
            await f.write(orjson.dumps(self._raw_member_info, indent=4))

    async def networth(
        self, include_banking_data_if_available: bool = True
    ) -> Dict[str, float]:
        """Gets the networth of a profile from the Maro API (https://nariah-dev.com/)"""
        body = {"data": self._raw_member_info, "success": True}
        if include_banking_data_if_available:
            if 'banking' in self._raw:
                body["data"]["banking"] = self._raw["banking"]
        async with self._hypy.session.post(
            "https://nariah-dev.com/api/networth/total",
            json=body,
            headers={"uuid": self._uuid},
        ) as response:
            jsn = await response.json()
            if not str(jsn["status"]).startswith("2"):
                raise MaroNoSuccess(jsn["status"], jsn["cause"])
            return jsn["data"]

    @property
    def senither_weight(self) -> dict:
        """Returns weight in the Senither weight system"""
        senither = SenitherWeight(self)
        return senither.calc_weight()

    @property
    def skills(self) -> SkyblockSkills:
        """Skill levels"""
        return SkyblockSkills(self._raw_member_info, self._hypy, xp=False)

    @property
    def xp(self) -> SkyblockSkills:
        """Skill xp"""
        return SkyblockSkills(self._raw_member_info, self._hypy, xp=True)

    @property
    def sea_creature_kills(self) -> int:
        """Sea creature kills of profile"""
        kills = self.kills
        sc_kills = 0
        for sc in scs:
            sc_kills += kills[sc]
        return int(sc_kills)

    @property
    def slayer(self) -> SkyblockSlayers:
        """Slayers"""
        return SkyblockSlayers(self._raw_member_info)

    @property
    def inventory(self) -> SkyblockInventory:
        """Inventory data"""
        return SkyblockInventory(self._raw_member_info["inv_contents"], "INVENTORY")

    @property
    def backpacks(self) -> SkyblockBackpacks:
        """Skyblock Backpacks"""
        return SkyblockBackpacks(self._raw_member_info["backpack_contents"])

    @property
    def quiver(self) -> SkyblockInventory:
        """Skyblock Quiver"""
        return SkyblockInventory(self._raw_member_info["quiver"], "QUIVER")

    @property
    def ender_chest(self) -> SkyblockInventory:
        """Skyblock Enderchest"""
        return SkyblockInventory(
            self._raw_member_info["ender_chest_contents"], "ENDERCHEST"
        )

    @property
    def personal_vault(self) -> SkyblockInventory:
        """Skyblock Personal Vault"""
        return SkyblockInventory(
            self._raw_member_info["personal_vault_contents"], "PERSONAL_VAULT"
        )

    @property
    def candy_bag(self) -> SkyblockInventory:
        """Skyblock Candy Bag"""
        return SkyblockInventory(
            self._raw_member_info["candy_inventory_contents"], "CANDY_BAG"
        )

    @property
    def talisman_bag(self) -> SkyblockInventory:
        """Skyblock Talisman Bag"""
        return SkyblockInventory(self._raw_member_info["talisman_bag"], "TALISMAN_BAG")

    @property
    def fishing_bag(self) -> SkyblockInventory:
        """Skyblock Fishing Bag"""
        return SkyblockInventory(self._raw_member_info["fishing_bag"], "FISHING_BAG")

    @property
    def potion_bag(self) -> SkyblockInventory:
        """Skyblock Potion Bag"""
        return SkyblockInventory(self._raw_member_info["potion_bag"], "POTION_BAG")

    @property
    def armor(self) -> SkyblockInventory:
        """Armor data"""
        return SkyblockInventory(self._raw_member_info["inv_armor"], "ARMOR")

    @property
    def kills(self) -> defaultdict:
        """Dictionary of mob kills"""
        tr = defaultdict(lambda: 0)
        tr["total"] = self._raw_member_info["stats"]["kills"]
        for key in self._raw_member_info["stats"]:
            keystr = str(key)
            if keystr.startswith("kills_"):
                tr[keystr.replace("kills_", "")] = self._raw_member_info["stats"][
                    keystr
                ]
        return tr

    @property
    def deaths(self) -> defaultdict:
        """Dictionary of deaths"""
        tr = defaultdict(lambda: 0)
        tr["total"] = self._raw_member_info["stats"]["deaths"]
        for key in self._raw_member_info["stats"]:
            keystr = str(key)
            if keystr.startswith("deaths_"):
                tr[keystr.replace("deaths_", "")] = self._raw_member_info["stats"][
                    keystr
                ]
        return tr

    @property
    def purse(self) -> int:
        """Amount of coins in purse"""
        return self._raw_member_info["coin_purse"]

    @property
    def bank(self) -> SkyblockBank:
        """Coins in bank"""
        return SkyblockBank(self._raw)



from .hypyobject import HypyObject


class PlayerCounts(HypyObject):
    """Hypixel Player counts

    Get the counts of a game with playercounts.GAME_NAME
    """

    def __init__(self, raw) -> None:
        self._raw = raw

    def __int__(self) -> int:
        return self._raw["playerCount"]

    def __getattr__(self, name) -> int:
        return self._raw["games"][name.upper()]["players"]



from typing import Any, Optional
from .hypyobject import HypyObject

REGULAR_SKILLS = [
    "fishing",
    "alchemy",
    "taming",
    "enchanting",
    "combat",
    "mining",
    "farming",
    "foraging",
]

COSMETIC_SKILLS = [
    "carpentry",
    "runecrafting",
]

REGULAR_SKILL_FILTER = "experience_skill_{}"

DUNGEON_SKILLS = ["catacombs"]

CLASS_SKILLS = ["healer", "berserk", "archer", "mage", "tank"]

SLAYER_TYPES = ["zombie", "spider", "wolf", "enderman"]


class SkyblockSkills(HypyObject):
    """A class representing skyblock skills

    The level of a skill can be retrieved like this: skyblockskills.fishing
    """

    __slots__ = REGULAR_SKILLS + DUNGEON_SKILLS + CLASS_SKILLS

    def __init__(self, raw, hypy, xp=False) -> None:
        self._profile_info = raw
        self._hypy = hypy
        self._xp = xp

    @property
    def raw(self) -> dict:
        """Get skills in a dictionary style"""
        return {
            "detail": {
                "fishing": self.fishing,
                "alchemy": self.alchemy,
                "taming": self.taming,
                "enchanting": self.enchanting,
                "combat": self.combat,
                "mining": self.mining,
                "farming": self.farming,
                "foraging": self.foraging,
                "catacombs": self.catacombs,
            },
            "avg": self.avg,
            "classes": {
                "healer": self.healer,
                "berserk": self.berserk,
                "archer": self.archer,
                "mage": self.mage,
                "tank": self.tank,
            },
            "misc": {"carpentry": self.carpentry, "runecrafting": self.runecrafting},
        }

    @property
    def _raw(self):
        return self.raw

    def __getitem__(self, key) -> Any:
        return self._raw[key]

    def __getattr__(self, name) -> float:
        try:
            return self._get_skill_info(name)
        except KeyError:
            return 0.0

    def _get_skill_info(self, skill) -> float:
        if skill in REGULAR_SKILLS + COSMETIC_SKILLS:
            return float(self._get_regular_skill_info(skill))
        if skill in DUNGEON_SKILLS:
            return float(self._get_dungeon_skill_info(skill))
        if skill in CLASS_SKILLS:
            return float(self._get_class_skill_info(skill))
        if skill == "avg":
            return sum([self.__getattr__(x) for x in REGULAR_SKILLS]) / len(
                REGULAR_SKILLS
            )
        return 0.0

    def _get_regular_skill_info(self, skill) -> float:
        if REGULAR_SKILL_FILTER.format(skill) in self._profile_info:
            if self._xp:
                return self._profile_info[REGULAR_SKILL_FILTER.format(skill)]
            return self._hypy.utils.lvlCalc(
                self._profile_info[REGULAR_SKILL_FILTER.format(skill)], skill
            )
        return 0.0

    def _get_dungeon_skill_info(self, skill) -> float:
        if "dungeons" not in self._profile_info:
            return 0.0
        if skill not in self._profile_info["dungeons"]["dungeon_types"]:
            return 0.0
        if "experience" not in self._profile_info["dungeons"]["dungeon_types"][skill]:
            return 0.0
        if self._xp:
            return self._profile_info["dungeons"]["dungeon_types"][skill]["experience"]
        return self._hypy.utils.lvlCalc(
            self._profile_info["dungeons"]["dungeon_types"][skill]["experience"], skill
        )

    def _get_class_skill_info(self, skill) -> float:
        if "dungeons" not in self._profile_info:
            return 0.0
        if "player_classes" not in self._profile_info["dungeons"]:
            return 0.0
        if skill not in self._profile_info["dungeons"]["player_classes"]:
            return 0.0
        if "experience" not in self._profile_info["dungeons"]["player_classes"][skill]:
            return 0.0
        if self._xp:
            return self._profile_info["dungeons"]["player_classes"][skill]["experience"]
        return self._hypy.utils.lvlCalc(
            self._profile_info["dungeons"]["player_classes"][skill]["experience"],
            "catacombs",
        )


class SkyblockSlayers(HypyObject):
    """A class representing Skyblock Slayers of a profile

    The xp of a slayer can be retrieved like this: skyblockskills.enderman
    """

    def __init__(self, raw):
        self._profile_info = raw
        self._slayer_info = self._profile_info["slayer_bosses"]

    @property
    def raw(self) -> dict:
        """Get slayers in a dictionary style"""
        return {
            "detail": {
                "wolf": self.wolf,
                "enderman": self.enderman,
                "spider": self.spider,
                "zombie": self.zombie,
            },
            "total": self.total,
        }

    @property
    def _raw(self):
        return self.raw

    def __getitem__(self, key) -> int:
        return self._raw[key]

    def __getattr__(self, name) -> int:
        try:
            if name == "total":
                return sum([self.__getattr__(x) for x in SLAYER_TYPES])
            return self._get_slayer_info(name)
        except KeyError:
            return 0

    def _get_slayer_info(self, slayer_type):
        if slayer_type not in self._slayer_info:
            return 0
        if "xp" not in self._slayer_info[slayer_type]:
            return 0
        return self._slayer_info[slayer_type]["xp"]


class SkyblockBank(HypyObject):
    """A class representing the skyblock bank of a profile"""

    def __init__(self, raw) -> None:
        self._raw_data = raw
        self._api_on = "banking" in self._raw_data

    @property
    def _raw(self):
        return self.raw

    @property
    def raw(self) -> dict:
        """A dict style representation of SkyblockBank"""
        return {
            "balance": self.balance or 0,
            "msg": "-" if self._api_on else "API is off",
            "success": self._api_on,
        }

    def __getitem__(self, key):
        return self.raw[key]

    @property
    def balance(self) -> Optional[int]:
        """Balance in bank"""
        if self._api_on:
            return self._raw_data["banking"]["balance"]
        return None



import asyncio
from typing import Iterator, List
from .hypyobject import HypyObject
from .abc import HypyIterable
from .hypixelfriend import HypixelFriend
from .uuid import UUID


class HypixelFriends(HypyObject, HypyIterable):
    """Hypixel Friends of a Player"""

    __slots__ = ("_raw", "_hypy")

    friends: List[HypixelFriend]
    """A list of HypixelFriends"""

    def __init__(self, raw, hypy) -> None:
        self._raw = raw
        self._hypy = hypy
        self.of = UUID(self._raw["uuid"])
        self.friends = [
            HypixelFriend(friend, self.of, hypy) for friend in self._raw["records"]
        ]

    def iter_uuids(self) -> Iterator[UUID]:
        for i in self:
            yield i.uuid

    def __iter__(self) -> Iterator[HypixelFriend]:
        for i in self.friends:
            yield i

    async def names(self) -> List[str]:
        """Return names of all of the friends"""
        return await asyncio.gather(
            *[friend.name() for friend in sorted(self.friends, key=lambda f: f.since)]
        )



from abc import ABC
from typing import Iterator
from .uuid import UUID


class HypyIterable(ABC):
    """A hypy Iterable"""

    def iter_uuids(self) -> Iterator[UUID]:
        """Iterate over uuids of Iterable"""


class MinecraftPlayer(ABC):
    """A minecraft player"""

    uuid: UUID



from typing import List
from .hypyobject import HypyObject


class BazaarOrder(HypyObject):
    """An order on the SkyBlock Bazaar"""

    type: str
    """Whether the order is a BUY or SELL order"""
    price_per_unit: int
    """The price per item"""
    amount: int
    """The amount of items ordered"""

    def __init__(self, raw: dict, order_type: str, item: str) -> None:
        self._raw = raw
        self.type = order_type
        self.item = item
        self.price_per_unit = self._raw["pricePerUnit"]
        self.orders = self._raw["orders"]
        self.amount = self._raw["amount"]

    def __repr__(self) -> str:
        return f"<hypy.BazaarOrder item={self.item} type={self.type} price_per_unit={self.price_per_unit} amount={self.amount}>"


class BazaarItem(HypyObject):
    """An item on the SkyBlock Bazaar"""

    product_id: str
    """The items id"""
    buy_orders: List[BazaarOrder]
    """A list of buy orders for the item"""
    sell_orders: List[BazaarOrder]
    """A list of sell orders for the item"""

    def __init__(self, raw: dict) -> None:
        self._raw = raw
        self.product_id = self._raw["product_id"]
        self.buy_orders = [
            BazaarOrder(data, "BUY", self.product_id)
            for data in self._raw["buy_summary"]
        ]
        self.buy_orders = [
            BazaarOrder(data, "SELL", self.product_id)
            for data in self._raw["sell_summary"]
        ]

    def __repr__(self) -> str:
        return f"<hypy.BazaarItem item={self.product_id}>"



from typing import Optional
from .hypyobject import HypyObject


class PlayerStatus(HypyObject):
    """Status of a player"""

    online: bool
    """Whether the player is currently online"""
    game: Optional[str]
    """The game the player is currently playing"""
    mode: Optional[str]
    """The mode of the game the player is currently playing"""
    map: Optional[str]
    """The map the player is currently playing on"""

    def __init__(self, data) -> None:
        self._raw = data["session"]
        self.online = self._raw["online"]
        self.game = self._raw["gameType"] if self.online else None
        self.mode = self._raw["mode"] if self.online else None
        self.map = self._raw["map"] if ("map" in self._raw) else None



from typing import Tuple, Optional
import asyncio
import traceback
from time import time_ns as timestamp
from asyncio import sleep
from urllib.parse import quote
from orjson import orjson
import aiohttp
from .player import Player
from .mojang import Mojang
from .playerstatus import PlayerStatus
from .exceptions import (
    ContentTypeException,
    MissingParamException,
    HypixelNoSuccess,
    ExceededMaxRetries,
    InvalidAccountDescriptor,
    NotEnoughArguments,
)
from .guild import Guild
from .hypixelresources import SkyBlockResources
from .auctions import SkyblockAuctions
from .auction import multi_init
from . import utils
from .utils import Utils
from .models import KeyStats, WatchdogStats
from .profile import SkyblockProfile
from .bazaar import Bazaar
from .hypixelfriends import HypixelFriends
from .playercounts import PlayerCounts


class Hypixel:
    """The main object used to interact with the Hypixel API

    Before interacting with the api you need to call Hypixel.setup()
    """

    _apikey = ""
    loop: asyncio.AbstractEventLoop
    session: aiohttp.ClientSession
    _base_url = "https://api.hypixel.net/"
    mojang: Mojang
    _sb_resources = None
    utils = None
    _debug = False
    __version__ = "1.3.1"
    _retry = False
    _max_retries = 0
    _total_calls = 0

    def __init__(
        self,
        key,
        *,
        session: aiohttp.ClientSession = None,
        debug: bool = False,
        retry: bool = False,
        max_retries: int = 5,
        loop: asyncio.AbstractEventLoop = None,
    ):
        if session is not None:
            self.session = session
        self._apikey = key
        self._debug = debug
        self._retry = retry
        self._max_retries = max_retries
        self._headers = {"API-Key": self._apikey}
        self.loop = loop or asyncio.get_event_loop()

    async def close(self) -> None:
        """Close internal session"""
        await self.session.close()

    async def setup(self) -> None:
        """Properly set up Hypixel object"""
        if not hasattr(self, "session"):
            self.session = aiohttp.ClientSession(loop=self.loop)
        self.mojang = Mojang(session=self.session)
        await self.update_resources()
        self.utils = Utils(self._sb_resources)

    def _debug_url(self, endpoint, **params) -> str:
        """Generate a debug url for endpoint and parameters"""
        url = f"{self._base_url}{endpoint}?key={self._apikey}"
        for key, value in params.items():
            if key == "_HYPY_RETRIES":
                continue
            url += f"&{key}={quote(value)}"
        return url

    async def _get(self, endpoint, **params) -> Tuple[int, dict]:
        """GET something from the Hypixel API"""
        self._total_calls += 1
        endpoint = endpoint.lstrip("/")  # is this even needed
        if self._debug:
            print(self._debug_url(endpoint, **params))
        url = f"{self._base_url}{endpoint}?"
        retries = 0
        if "_HYPY_RETRIES" in params:
            retries = int(params["_HYPY_RETRIES"])
            del params["_HYPY_RETRIES"]
        for key, value in params.items():
            url += f"{key}={quote(value)}&"
        url = url[:-1]
        async with self.session.get(url, headers=self._headers) as res:
            try:
                jsn = await res.json(loads=orjson.loads)
                if not jsn["success"]:
                    raise HypixelNoSuccess(jsn["cause"])
                return res.status, jsn
            except aiohttp.client_exceptions.ContentTypeError:
                if self._debug:
                    print("In contenttype handler")
                if not self._retry:
                    raise ContentTypeException(res.headers.get("content-type"))
                if retries > self._max_retries:
                    raise ExceededMaxRetries(self._max_retries)
                if self._debug:
                    print("RETRYING {} RETRY NUMBER {}".format(url, retries + 1))
                rrq = await self._get(
                    endpoint, **dict(**params, **{"_HYPY_RETRIES": retries + 1})
                )
                return rrq[0], rrq[1]
            except aiohttp.ClientResponseError:
                if self._debug:
                    traceback.print_exc()
                raise

    async def get_player(self, nameOrUuid: str) -> Player:
        """Returns a Player object for a given name or UUID

        :param nameOrUuid: The name or uuid of the player
        """
        if utils.is_username(nameOrUuid):
            nameOrUuid = await self.mojang.name_to_uuid(nameOrUuid)
        elif not utils.is_uuid(nameOrUuid):
            raise InvalidAccountDescriptor(nameOrUuid)

        _, response = await self._get("/player", uuid=nameOrUuid)
        return Player(response, nameOrUuid, self)

    async def get_player_status(self, nameOrUuid: str) -> PlayerStatus:
        """Returns status of player

        :param nameOrUuid: The name or uuid of the player
        """
        if utils.is_username(nameOrUuid):
            nameOrUuid = await self.mojang.name_to_uuid(nameOrUuid)
        elif not utils.is_uuid(nameOrUuid):
            raise InvalidAccountDescriptor(nameOrUuid)

        _, response = await self._get("/status", uuid=nameOrUuid)
        return PlayerStatus(response)

    async def get_guild(
        self, name: str = None, guild_id: str = None, playerNameOrUuid: str = None
    ) -> Optional[Guild]:
        """Gets a guild based on one of three parameters

        :param name: The name of a guild
        :param guild_id: The id of a guild
        :param playerNameOrUuid: The name or uuid of a guild member
        """
        if name is not None:
            _, response = await self._get("/guild", name=name)
        elif guild_id is not None:
            _, response = await self._get("/guild", id=guild_id)
        elif playerNameOrUuid is not None:
            if utils.is_username(playerNameOrUuid):
                playerNameOrUuid = await self.mojang.name_to_uuid(playerNameOrUuid)
            elif not utils.is_uuid(playerNameOrUuid):
                raise InvalidAccountDescriptor(playerNameOrUuid)
            _, response = await self._get("/guild", player=playerNameOrUuid)
        else:
            raise NotEnoughArguments("name, id, playerNameOrUuid")

        if response["guild"] is not None:
            return Guild(response, self)
        return None

    async def get_key_stats(self) -> KeyStats:
        """Gets statistics of API Key"""
        _, res = await self._get("/key")
        res = res["record"]
        return KeyStats(
            res["owner"],
            await self.mojang.uuid_to_name(res["owner"]),
            res["totalQueries"],
            res["queriesInPastMin"],
            timestamp(),
        )

    async def get_watchdog_stats(self) -> WatchdogStats:
        """Gets punishment statistics"""
        _, res = await self._get("punishmentstats")
        return WatchdogStats(
            res["watchdog_lastMinute"],
            res["staff_rollingDaily"],
            res["watchdog_total"],
            res["watchdog_rollingDaily"],
            res["staff_total"],
        )

    async def get_profile(self, profile_id: str, uuid: str) -> SkyblockProfile:
        """Gets a Hypixel SkyBlock Profile

        For general usage, going through :py:func:`hypy.player.Player.find_profile` is better

        :param profile_id: The id of the profile
        :param uuid: The uuid of the player
        """
        if not profile_id:
            raise MissingParamException("name, profile_id")
        _, raw_profile_data = await self._get("skyblock/profile", profile=profile_id)

        return SkyblockProfile(raw_profile_data, uuid, self)

    async def get_profile_data(self, profile_id) -> dict:
        """Gets Hypixel SkyBlock Profile data based on profile id"""
        if not profile_id:
            raise MissingParamException("name, profile_id")
        _, raw_profile_data = await self._get("skyblock/profile", profile=profile_id)

        return raw_profile_data

    async def get_friends(self, nameOrUuid: str) -> HypixelFriends:
        """Gets Friends of given name or UUID

        Args:
            nameOrUuid: The name or uuid of the player
        """
        if utils.is_username(nameOrUuid):
            nameOrUuid = await self.mojang.name_to_uuid(nameOrUuid)
        elif not utils.is_uuid(nameOrUuid):
            raise InvalidAccountDescriptor(nameOrUuid)

        _, response = await self._get("/friends", uuid=nameOrUuid)
        return HypixelFriends(response, self)

    async def get_auctions(self) -> SkyblockAuctions:
        """Gets Hypixel SkyBlock auctions"""
        all_itime_equal = False
        while not all_itime_equal:
            _, init_response = await self._get("/skyblock/auctions")
            i_time = init_response["lastUpdated"]
            i_pages = init_response["totalPages"]
            results = [
                x[1]
                for x in list(
                    await asyncio.gather(
                        *[
                            self._get("skyblock/auctions", page=str(pn))
                            for pn in range(1, i_pages)
                        ]
                    )
                )
            ]
            results.insert(0, init_response)
            all_itime_equal = all([i_time == x["lastUpdated"] for x in results])
            if not all_itime_equal:
                await sleep(10)  # i think this is good to do regardless
        return SkyblockAuctions(
            *await asyncio.to_thread(multi_init, results, self), self
        )

    async def find_guild(self, nameOrUuid) -> Guild:
        """Deprecated, use Hypixel.getGuild instead"""
        return await self.get_guild(playerNameOrUuid=nameOrUuid)

    async def get_bazaar(self) -> Bazaar:
        """Gets bazaar information"""
        _, response = await self._get("/skyblock/bazaar")
        return Bazaar(response)

    async def get_player_counts(self) -> PlayerCounts:
        """Get Hypixel Player counts"""
        _, response = await self._get("/counts")
        return PlayerCounts(response)

    async def update_resources(self) -> None:
        """Update Hypixel Resources"""
        _, response = await self._get("/resources/skyblock/skills")
        self._sb_resources = SkyBlockResources(response)



from typing import Iterator, List, Dict
from .utils import unpack_nbt, decode_nbt, parse_timestamp


class SkyblockItemSlot:
    """A Skyblock Item Slot"""

    empty: bool = False
    """Whether the slot is empty"""
    enchantments: Dict[str, int] = {}
    """A dictionary of enchantments (name:level)"""
    origin: str = ""
    """How the item was obtained"""
    name: str = ""
    """The name of the item"""
    lore: str = ""
    """The lore of the item"""
    skyblock_item_id: str = ""
    """The skyblock item id of the item"""
    minecraft_item_id: int = 0
    """The minecraft item id of the item"""
    reforge: str = ""
    """The reforge of the item"""
    recombobulated: bool = False
    """Whether the items rarity has been upgraded"""
    count: int = 0
    """How many items there are"""

    def __init__(self, raw) -> None:
        self._raw = raw
        if not bool(self._raw):
            self.empty = True
            return
        self.minecraft_item_id = self._raw["id"]
        self.count = self._raw["Count"]
        self.damage = self._raw["Damage"]
        self.name = self._raw["tag"]["display"]["Name"]
        self.lore = self._raw["tag"]["display"]["Lore"]
        self.skyblock_item_id = self._raw["tag"]["ExtraAttributes"].get("id")
        self.uuid = self._raw["tag"]["ExtraAttributes"].get("uuid")
        self.reforge = self._raw["tag"]["ExtraAttributes"].get("modifier")
        self.recombobulated = bool(
            self._raw["tag"]["ExtraAttributes"].get("rarity_upgrades", False)
        )
        self.enchantments = self._raw["tag"]["ExtraAttributes"].get("enchantments")
        self.origin = self._raw["tag"]["ExtraAttributes"].get("originTag")
        self.timestamp = "timestamp" in self._raw["tag"][
            "ExtraAttributes"
        ] and parse_timestamp(self._raw["tag"]["ExtraAttributes"]["timestamp"])

    def __repr__(self) -> str:
        return f"<hypy.SkyBlockItemSlot empty={self.empty} skyblock_item_id={self.skyblock_item_id}>"

    def __str__(self) -> str:
        return self.__repr__()


class SkyblockInventory:
    """A Skyblock Inventory"""

    type: str
    """The type of the inventory"""
    slots: List[SkyblockItemSlot]
    """The list of items in the inventory"""

    def __init__(self, raw, inventory_type) -> None:
        self._raw = raw
        self.type = inventory_type
        self.nbt_data = unpack_nbt(decode_nbt(self._raw["data"]))["i"]
        self.slots = [SkyblockItemSlot(x) for x in self.nbt_data]

    def __repr__(self) -> str:
        return f"<hypy.SkyblockInventory type={self.type} slot_count={len(self)}>"

    def __len__(self) -> int:
        return len(self.slots)

    def __getitem__(self, key) -> SkyblockItemSlot:
        return self.slots[key]

    def __iter__(self) -> Iterator[SkyblockItemSlot]:
        for i in self.slots:
            yield i


class SkyblockBackpack(SkyblockInventory):
    """A Skyblock Backpack"""

    def __init__(self, raw, inventory_type, backpack_index) -> None:
        super().__init__(raw, "BACKPACK")
        self.backpack_index = backpack_index


class SkyblockBackpacks:
    """A list of SkyBlock backpacks"""

    backpacks: List[SkyblockBackpack]
    """A list of SkyblockBackpacks"""

    def __init__(self, raw) -> None:
        self._raw = raw
        self.backpacks = [
            SkyblockBackpack(v, "BACKPACK", int(k))
            for k, v in sorted(self._raw.items(), key=lambda x: int(x[0]))
        ]

    def __len__(self) -> int:
        return len(self.backpacks)

    def __getitem__(self, key) -> SkyblockBackpack:
        return self.backpacks[key]

    def __iter__(self) -> Iterator[SkyblockBackpack]:
        for i in self.backpacks:
            yield i



from __future__ import annotations
from typing import TYPE_CHECKING, List, Tuple
from datetime import datetime
from itertools import chain
from .uuid import UUID
from .hypyobject import HypyObject
from .utils import unpack_nbt, decode_nbt, get_auction_lores
from .skyblockinventories import SkyblockItemSlot

if TYPE_CHECKING:
    from .hypixel import Hypixel

USE_NBT = False


def multi_init(pages: List[dict], hypy: Hypixel) -> Tuple[List[dict], List[Auction]]:
    """initialize multiple auctions at once"""
    auctions = list(
        chain.from_iterable([x["auctions"] for x in pages])
    )  # join all auctions from all pages into one list
    lores = get_auction_lores(auctions)
    return (
        auctions,
        [Auction.init(info[0], info[1], hypy) for info in zip(auctions, lores)],
    )


class Auction(HypyObject):
    """A Hypixel SkyBlock Auction"""

    _uuid: str
    _raw: dict
    _hypy: Hypixel
    auctioneer_uuid: UUID
    """The UUID of the auctioneer"""
    profile_id: str
    """The profile id of the auctioneer"""
    _start: int
    _end: int
    item_name: str
    """The name of the item"""
    item_lore: str
    """The lore of the item
    If you want cleaned lore (without colorcodes) use `safe_item_lore` instead
    """
    safe_item_lore: str
    """The lore without any formatting"""
    item_category: str
    """The auction house category of the item"""
    tier: str
    """The rarity of the item"""
    starting_bid: str
    """The starting bid of the item"""
    _bytes: str
    bin: bool
    """Whether the auction is BIN (Buy it now) or not"""
    highest_bid: int
    """The highest bid of the auction"""

    def __init__(self, data, hypy) -> None:
        self._hypy = hypy
        self._raw = data
        self._uuid = data["uuid"]
        self.auctioneer_uuid = UUID(data["auctioneer"])
        self._start = data["start"]
        self._end = data["end"]
        self._bytes = data["item_bytes"]

    @staticmethod
    def init(data, lore, hypy) -> Auction:
        """Initialize an Auction object"""
        auc = Auction(data, hypy)
        auc.profile_id = data["profile_id"]
        auc.item_name = data["item_name"]
        auc.item_lore = data["item_lore"]
        auc.safe_item_lore = lore
        auc.item_category = data["category"]
        auc.tier = data["tier"]
        auc.starting_bid = data["starting_bid"]
        auc.bin = data["bin"] if ("bin" in data) else False
        auc.highest_bid = data["highest_bid_amount"]
        return auc

    def __str__(self) -> str:
        return f"<hypy.Auction item_name={self.item_name}>"

    @property
    def nbt_item(self) -> SkyblockItemSlot:
        """A SkyblockItemSlot for this auction"""
        return SkyblockItemSlot(unpack_nbt(decode_nbt(self._raw["item_bytes"]))["i"][0])

    @property
    def start(self) -> datetime:
        """The start of the auction"""
        return datetime.utcfromtimestamp(self._start / 1000)

    @property
    def end(self) -> datetime:
        """The end of the auction"""
        return datetime.utcfromtimestamp(self._end / 1000)

    async def get_auctioneer_name(self) -> str:
        """Get name of Auctioneer

        Returns:
            The name of the auctioneer

        Raises:
            UUIDNotFound: The uuid to find the name was invalid
        """
        return await self._hypy.mojang.uuid_to_name(self.auctioneer_uuid.no_dashes)



from typing import Optional


class SocialMedia:
    """Hypixel Social Media response wrapper"""

    def __init__(self, raw):
        self._raw = raw

    def _get_social(self, social) -> Optional[str]:
        if "socialMedia" not in self._raw:
            return None
        if "links" not in self._raw["socialMedia"]:
            return None
        if social not in self._raw["socialMedia"]["links"]:
            return None
        return str(self._raw["socialMedia"]["links"][social]).strip()

    @property
    def youtube(self) -> Optional[str]:
        """The linked youtube channel"""
        return self._get_social("YOUTUBE")

    @property
    def discord(self) -> Optional[str]:
        """The linked discord tag"""
        return self._get_social("DISCORD")

    @property
    def hypixel(self) -> Optional[str]:
        """The linked hypixel forums account"""
        return self._get_social("HYPIXEL")

    @property
    def mixer(self) -> Optional[str]:
        """The linked mixer account"""
        # did you know you could link mixer? cause i sure didnt.
        return self._get_social("MIXER")

    @property
    def twitch(self) -> Optional[str]:
        """The linked twitch account"""
        return self._get_social("TWITCH")

    @property
    def twitter(self) -> Optional[str]:
        """The linked twitter account"""
        return self._get_social("TWITTER")

    @property
    def instagram(self) -> Optional[str]:
        """The linked instagram account"""
        return self._get_social("INSTAGRAM")



import uuid as uuidc


class UUID:
    """Helper Class for uuid"""

    def __init__(self, uuid):
        self.uuid = uuidc.UUID(uuid)

    @property
    def dashes(self) -> str:
        """The UUID with dashes"""
        return str(self.uuid)

    @property
    def no_dashes(self) -> str:
        """The UUID without dashes"""
        return str(self.uuid).replace("-", "")

    def __str__(self) -> str:
        return self.no_dashes

    def __eq__(self, other) -> bool:
        return str(self) == str(UUID(other))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return f"<hypy.UUID uuid={str(self.no_dashes)}>"



from datetime import datetime
from .hypyobject import HypyObject


class GuildRank(HypyObject):
    """A Hypixel guild rank"""

    name: str
    """The name of the rank"""
    default: bool
    """Whether the rank is the default rank for the guild"""
    tag: str
    """The tag of the rank"""
    priority: int
    """The priority/position of the rank"""

    def __init__(self, raw, idx) -> None:
        self._raw = raw
        self._raw["inGuild"] = True  # backwards compat
        self._raw["rank"] = idx  # backwards compat
        self.name = self._raw["name"]
        self.default = bool(self._raw.get("default", False))
        self.tag = self._raw.get("tag", "")
        self.priority = self._raw.get("priority", 0)
        self.index = idx

    def __getitem__(self, key):
        return self._raw[key]

    def __repr__(self) -> str:
        return f'<hypy.GuildRank name="{self.name}" priority={self.priority} default={self.default} index={self.index}>'

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def created(self) -> datetime:
        """When the rank was created"""
        return datetime.utcfromtimestamp(self._raw["created"] / 1000)



from typing import Callable, Optional, List, Tuple
from datetime import datetime
from .exceptions import (
    MissingParamException,
    InvalidProfileNameException,
    InvalidProfileException,
    NoProfileInfoAvailableException,
    NoProfilesFoundException,
)
from .profile import SkyblockProfile
from .vars import skills, types, cute_names
from .hypyobject import HypyObject
from .socialmedia import SocialMedia
from .hypixelplayer import HypixelPlayer


class Player(HypyObject, HypixelPlayer):
    """A Hypixel Player"""

    __slots__ = ("_raw", "_hypy", "_uuid")

    displayname: str
    """The display name of the player"""

    def __init__(self, data, uuid, hypy) -> None:
        self._hypy = hypy
        self._uuid = self.uuid = uuid
        self._raw = data = data["player"]
        if "displayname" in data:
            self.displayname = data["displayname"]
        else:
            self.displayname = data["playername"]

    def __repr__(self) -> str:
        return f"<hypy.Player {self.displayname} last_login={self.last_login}>"

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def first_login(self) -> datetime:
        """The first login of the Player"""
        return datetime.utcfromtimestamp(self._raw["firstLogin"] / 1000)

    @property
    def last_login(self) -> datetime:
        """The last login of the Player"""
        return datetime.utcfromtimestamp(self._raw["lastLogin"] / 1000)

    @property
    def last_logout(self) -> datetime:
        """The last logout of the Player"""
        return datetime.utcfromtimestamp(self._raw["lastLogout"] / 1000)

    @property
    def social_media(self) -> SocialMedia:
        """The players social media information"""
        return SocialMedia(self._raw)

    @property
    def achievement_skills(self) -> dict:
        """SkyBlock skill data from achievements"""
        rt: dict = {"detail": {}, "avg": 0}
        for skill in skills:
            qdic = self._raw["achievements"]
            qstr = f"skyblock_{types[skill]}"
            if qstr not in qdic:
                rt["detail"][skill] = 0
                continue
            if skill in ["catacombs"]:
                rt["detail"][skill] = qdic[qstr]
            else:
                lvl = qdic[qstr]
                rt["detail"][skill] = lvl
                rt["avg"] += lvl
        rt["avg"] /= 8
        return rt

    @property
    def skyblock_profiles(self) -> List[dict]:
        """SkyBlock Profiles of the Player"""
        plist = []
        for _, value in self._raw["stats"]["SkyBlock"]["profiles"].items():
            plist.append(value)
        return plist

    async def get_profile(self, *, name=None, profile_id=None) -> SkyblockProfile:
        """Gets a Hypixel SkyBlock Profile from name OR profile id

        :param name: The cute name (eg. Strawberry) of a profile
        :param profile_id: The profile_id of a profile
        """
        if not profile_id:
            if not name:
                raise MissingParamException("name, profile_id")
            if name not in cute_names:
                raise InvalidProfileNameException(name)
            pl = [
                (x["profile_id"], x["cute_name"])
                for x in self.skyblock_profiles
                if x["cute_name"] == str.capitalize(str.lower(name))
            ]
            if len(pl) == 0:
                raise InvalidProfileException(f"{name} {profile_id}")
            profile_id = pl[0][0]
            cute_name = pl[0][1]

        profile = await self._hypy.get_profile(profile_id, self._uuid)
        profile.add_cute_name(cute_name)
        return profile

    async def find_profile(
        self, *, key: Callable
    ) -> Tuple[SkyblockProfile, Optional[str]]:
        """Find Hypixel SkyBlock Profile by a metric

        :param key: The key to find the profile by, most likely a lambda (eg. find_profile(key=lambda p: p.skills.avg))
        """
        profiles = [
            await self.get_profile(name=profile["cute_name"])
            for profile in self.skyblock_profiles
        ]
        profiles = [p for p in profiles if not p.deleted]
        if len(profiles) == 0:
            raise NoProfileInfoAvailableException()
        match = sorted(profiles, key=key, reverse=True)[0]
        if match is not None:
            return match, match.cute_name
        raise NoProfilesFoundException(self._uuid)



from typing import Set
from ...abc import HypyIterable


def player_collisions(party1: HypyIterable, party2: HypyIterable) -> Set[str]:
    """Returns collisions of people between 2 Objects"""
    parties = []
    for i in [party1, party2]:
        parties.append([x.no_dashes for x in i.iter_uuids()])
    return set(set(parties[0]).intersection(parties[1]))



from datetime import datetime
from .hypyobject import HypyObject
from .hypixelplayer import HypixelPlayer


class GuildMember(HypixelPlayer, HypyObject):
    """A Hypixel Guild Member"""

    def __init__(self, raw, hypy) -> None:
        self._raw = raw
        self._hypy = hypy
        self.uuid = self._raw["uuid"]
        self.rank = self._raw["rank"]
        self.quest_participation = int(self._raw.get("questParticipation") or 0)
        self.exp_history = self._raw["expHistory"]

    def __getitem__(self, key):
        return self._raw[key]

    @property
    def joined(self) -> datetime:
        """The datetime when the player joined the guild"""
        return datetime.utcfromtimestamp(self._raw["joined"] / 1000)



from .hypyobject import HypyObject
from .player import Player
from .hypixel import Hypixel
from .guild import Guild
from .profile import SkyblockProfile
from .mojang import Mojang
from .exceptions import *
from .auctions import FilterType, SkyblockAuctions
from .auction import Auction
from .uuid import UUID
from .bazaar import Bazaar
from .bazaaritem import BazaarItem, BazaarOrder
from .hypixelfriend import HypixelFriend
from .hypixelfriends import HypixelFriends
from . import utils



from typing import Iterator
from datetime import datetime
from .hypyobject import HypyObject
from .bazaaritem import BazaarItem


class Bazaar(HypyObject):
    """A class containing information about the Hypixel Skyblock bazaar

    Items can be accessed with Bazaar['ITEM_NAME']
    """

    def __init__(self, raw) -> None:
        self._raw = raw
        self._timestamp = self._raw["lastUpdated"]
        self._items = [
            BazaarItem(item_data) for item_data in self._raw["products"].values()
        ]

    def __getitem__(self, key) -> BazaarItem:
        return self._items[key]

    def __iter__(self) -> Iterator[BazaarItem]:
        for i in self._items:
            yield i

    @property
    def timestamp(self) -> datetime:
        """The time when Bazaar information was fetched"""
        return datetime.utcfromtimestamp(self._timestamp / 1000)



from dataclasses import dataclass


@dataclass
class KeyStats:
    """API Key Statistics"""

    owner: str
    """The owner of the api key"""
    owner_name: str
    """The name of the owner of the api key"""
    total_uses: int
    """The total uses of the api key"""
    queries_in_past_minute: int
    """The amount of queries of the api key in the past minute"""
    timestamp: int
    """The timestamp of this information"""


@dataclass
class WatchdogStats:
    """Punishment statistics"""

    watchdog_last_minute: int
    """The amount of bans made by watchdog in the last minute"""
    staff_rolling_daily: int
    """The amount of bans made by staff in the past day"""
    watchdog_total: int
    """The total amount of bans made by watchdog"""
    watchdog_rolling_daily: int
    """The amount of bans made by watchdog in the past day"""
    staff_total: int
    """The total amount of bans made by watchdog"""



from __future__ import annotations
from typing import TYPE_CHECKING
from datetime import datetime
from .hypyobject import HypyObject
from .uuid import UUID
from .hypixelplayer import HypixelPlayer

if TYPE_CHECKING:
    from .player import Player
    from .hypixelfriends import HypixelFriends


class HypixelFriend(HypyObject, HypixelPlayer):
    """A Hypixel Friend"""

    __slots__ = ("_raw", "_hypy", "_id")

    def __init__(self, raw, of, _hypy) -> None:
        self._raw: dict = raw
        self._hypy = _hypy
        self.of = of
        self._id = self._raw["_id"]
        self.uuid = UUID(
            self._raw["uuidReceiver"]
            if of.no_dashes == self._raw["uuidSender"]
            else self._raw["uuidSender"]
        )

    @property
    def since(self) -> datetime:
        """Since when the friendship exists"""
        return datetime.utcfromtimestamp(self._raw["started"] / 1000)



import os
from setuptools import setup


def read(fname):
    """read... file"""
    # https://pythonhosted.org/an_example_pypi_project/setuptools.html
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="hypy-hypixel",
    version="1.3.1",
    description="A Hypixel API Wrapper",
    url="https://github.com/PikaFan123/hypy",
    author="PikaFan123",
    license="MIT License",
    long_description=read("readme.rst"),
    packages=["hypy", "hypy.ext", "hypy.ext.collisions", "hypy.ext.senither"],
    install_requires=["aiohttp", "dataclasses", "aiofiles", "nbt", "orjson"],
    python_requires=">=3.8,<4.0",
)
